"""Explicit Ansb transfer entropy estimator tests."""

import pytest
from numpy import nan, isnan

from infomeasure.estimators.transfer_entropy import (
    AnsbTEEstimator,
    AnsbCTEEstimator,
)


@pytest.mark.parametrize(
    "source,dest,base,expected",
    [
        ([1, 1, 1, 1, 1], [1, 1, 1, 1, 1], 2, 0.0),
        ([1, 0, 1, 0], [1, 0, 1, 0], 2, 0.0),
        ([1, 2, 3, 4, 5], [1, 2, 3, 4, 5], 2, nan),
        ([1, 0, 1, 0], [0, 1, 0, 1], 2, 0.0),
    ],
)
def test_ansb_te(source, dest, base, expected):
    """Test the ansb transfer entropy estimator."""
    est = AnsbTEEstimator(source, dest, base=base)
    res = est.result()
    if isnan(expected):
        assert isnan(res)
    else:
        assert res == pytest.approx(expected, rel=1e-10)


@pytest.mark.parametrize(
    "source,dest,cond,base,expected",
    [
        ([1, 1, 1, 1, 1], [1, 1, 1, 1, 1], [1, 1, 1, 1, 1], 2, 0.0),
        ([1, 0, 1, 0], [1, 0, 1, 0], [1, 0, 1, 0], 2, 0.0),
        ([1, 2, 3, 4, 5], [1, 2, 3, 4, 5], [1, 2, 3, 1, 5], 2, nan),
    ],
)
def test_ansb_cte(source, dest, cond, base, expected):
    """Test the ansb conditional transfer entropy estimator."""
    est = AnsbCTEEstimator(source, dest, cond=cond, base=base)
    res = est.result()
    if isnan(expected):
        assert isnan(res)
    else:
        assert res == pytest.approx(expected, rel=1e-10)



"""Module for the Nsb mutual information estimator."""

from abc import ABC

from numpy import issubdtype, integer

from infomeasure.estimators.base import (
    MutualInformationEstimator,
    ConditionalMutualInformationEstimator,
)

from ..entropy.nsb import NsbEntropyEstimator
from infomeasure import Config
from infomeasure.utils.types import LogBaseType


class BaseNsbMIEstimator(ABC):
    r"""Base class for Nsb mutual information estimators.

    Attributes
    ----------
    *data : array-like, shape (n_samples,)
        The data used to estimate the (conditional) mutual information.
        You can pass an arbitrary number of data arrays as positional arguments.
        For conditional mutual information, only two data arrays are allowed.
    cond : array-like, optional
        The conditional data used to estimate the conditional mutual information.
    K : int, optional
        The support size. If not provided, uses the observed support size.
    offset : int, optional
        Number of positions to shift the data arrays relative to each other.
        Delay/lag/shift between the variables. Default is no shift.
        Not compatible with the ``cond`` parameter / conditional MI.
    """

    def __init__(
        self,
        *data,
        cond=None,
        K: int = None,
        offset: int = 0,
        base: LogBaseType = Config.get("base"),
        **kwargs,
    ):
        r"""Initialize the Nsb estimator with specific parameters.

        Parameters
        ----------
        *data : array-like, shape (n_samples,)
            The data used to estimate the (conditional) mutual information.
            You can pass an arbitrary number of data arrays as positional arguments.
            For conditional mutual information, only two data arrays are allowed.
        cond : array-like, optional
            The conditional data used to estimate the conditional mutual information.
        K : int, optional
            The support size. If not provided, uses the observed support size.
        offset : int, optional
            Number of positions to shift the X and Y data arrays relative to each other.
            Delay/lag/shift between the variables. Default is no shift.
            Not compatible with the ``cond`` parameter / conditional MI.
        """
        if cond is None:
            super().__init__(*data, offset=offset, base=base, **kwargs)
        else:
            super().__init__(
                *data,
                cond=cond,
                offset=offset,
                base=base,
                **kwargs,
            )
        if K is not None and (not isinstance(K, int) or K <= 0):
            raise ValueError("The K parameter must be a positive integer.")

        self.K = K


class NsbMIEstimator(BaseNsbMIEstimator, MutualInformationEstimator):
    r"""Estimator for the Nsb mutual information.

    Nsb mutual information estimator using the entropy combination formula.

    Attributes
    ----------
    *data : array-like, shape (n_samples,)
        The data used to estimate the mutual information.
        You can pass an arbitrary number of data arrays as positional arguments.
    K : int, optional
        The support size. If not provided, uses the observed support size.
    offset : int, optional
        Number of positions to shift the data arrays relative to each other.
        Delay/lag/shift between the variables. Default is no shift.

    Notes
    -----
    This estimator uses the Nsb entropy estimator to compute mutual
    information through the entropy combination formula.

    Note that the entropy combination formula is used (_generic_mi_from_entropy)
    not a dedicated implementation as other MI might have.

    See Also
    --------
    infomeasure.estimators.entropy.nsb.NsbEntropyEstimator
        Nsb entropy estimator.
    """

    def _calculate(self):
        """Calculate the mutual information of the data.

        Returns
        -------
        float
            Nsb mutual information of the data.
        """

        return self._generic_mi_from_entropy(
            estimator=NsbEntropyEstimator,
            kwargs={"K": self.K, "base": self.base},
        )


class NsbCMIEstimator(BaseNsbMIEstimator, ConditionalMutualInformationEstimator):
    r"""Estimator for the conditional Nsb mutual information.

    Nsb conditional mutual information estimator using the entropy combination formula.

    Attributes
    ----------
    *data : array-like, shape (n_samples,)
        The data used to estimate the conditional mutual information.
        You can pass an arbitrary number of data arrays as positional arguments.
    cond : array-like
        The conditional data used to estimate the conditional mutual information.
    K : int, optional
        The support size. If not provided, uses the observed support size.
    offset : int, optional
        Number of positions to shift the data arrays relative to each other.
        Delay/lag/shift between the variables. Default is no shift.

    Notes
    -----
    This estimator uses the Nsb entropy estimator to compute conditional
    mutual information through the entropy combination formula.

    Note that the entropy combination formula is used (_generic_cmi_from_entropy)
    not a dedicated implementation as other MI might have.

    See Also
    --------
    infomeasure.estimators.entropy.nsb.NsbEntropyEstimator
        Nsb entropy estimator.
    """

    def _calculate(self):
        """Calculate the conditional mutual information of the data.

        Returns
        -------
        float
            Conditional Nsb mutual information of the data.
        """
        return self._generic_cmi_from_entropy(
            estimator=NsbEntropyEstimator,
            kwargs={"K": self.K, "base": self.base},
        )



"""Module for the discrete Miller-Madow mutual information estimator."""

from abc import ABC

from numpy import ndarray

from ..base import (
    MutualInformationEstimator,
    ConditionalMutualInformationEstimator,
)
from ..mixins import DiscreteMIMixin
from ..utils.discrete_interaction_information import (
    mutual_information_global,
    mutual_information_local,
    conditional_mutual_information_global,
    conditional_mutual_information_local,
)
from ... import Config
from ...utils.types import LogBaseType


class BaseMillerMadowMIEstimator(DiscreteMIMixin, ABC):
    """Base class for discrete Miller-Madow mutual information estimators.

    Attributes
    ----------
    *data : array-like, shape (n_samples,)
        The data used to estimate the (conditional) mutual information.
        You can pass an arbitrary number of data arrays as positional arguments.
        For conditional mutual information, only two data arrays are allowed.
    cond : array-like, optional
        The conditional data used to estimate the conditional mutual information.
    offset : int, optional
        Number of positions to shift the data arrays relative to each other.
        Delay/lag/shift between the variables. Default is no shift.
        Not compatible with the ``cond`` parameter / conditional MI.
    """

    def __init__(
        self,
        *data,
        cond=None,
        offset: int = 0,
        base: LogBaseType = Config.get("base"),
        **kwargs,
    ):
        """Initialize the BaseMillerMadowMIEstimator.

        Parameters
        ----------
        *data : array-like, shape (n_samples,)
            The data used to estimate the (conditional) mutual information.
            You can pass an arbitrary number of data arrays as positional arguments.
            For conditional mutual information, only two data arrays are allowed.
        cond : array-like, optional
            The conditional data used to estimate the conditional mutual information.
        offset : int, optional
            Number of positions to shift the X and Y data arrays relative to each other.
            Delay/lag/shift between the variables. Default is no shift.
            Not compatible with the ``cond`` parameter / conditional MI.
        """
        self.data: tuple[ndarray] = None
        if cond is None:
            super().__init__(*data, offset=offset, normalize=False, base=base, **kwargs)
        else:
            super().__init__(
                *data, cond=cond, offset=offset, normalize=False, base=base, **kwargs
            )
        self._check_data_mi()


class MillerMadowMIEstimator(BaseMillerMadowMIEstimator, MutualInformationEstimator):
    r"""Estimator for the discrete Miller-Madow mutual information.

    .. math::

        \begin{align}\hat{I}_{\tiny{MM}}(X; Y)
        &= \hat{H}_{\tiny{MM}}(X) + \hat{H}_{\tiny{MM}}(Y) - \hat{H}_{\tiny{MM}}(X, Y)\\
        &= \hat{H}_{\tiny{MLE}}(X) + \hat{H}_{\tiny{MLE}}(Y) - \hat{H}_{\tiny{MLE}}(X, Y)
          + (K_X + K_Y - K_{XY} - 1)/(2N \cdot \log(\texttt{base}))\\
        &= \hat{I}_{\tiny{MLE}}(X; Y) + (K_X + K_Y - K_{XY} - 1)/(2N \cdot \log(\texttt{base}))\\
        \end{align}

    For an arbitrary number of random variables this is equivalent to:

    .. math::

        \begin{align}\hat{I}_{\tiny{MM}}(X_1; \dots; X_n)
        &= \hat{I}_{\tiny{MLE}}(X_1; \dots; X_n)
        + \frac{\left(\sum_{i=1}^{n}K_i-1\right) - \left(K_{1,\dots,i}-1\right)}{2N \cdot \log(\texttt{base})}
        \end{align}

    :math:`\hat{I}_{\tiny{MLE}}(X_1; \dots; X_n)` is the
    initial :class:`~infomeasure.estimators.mutual_information.discrete.DiscreteMIEstimator` estimate,
    :math:`K_i` is the number of unique values in the i-th variable,
    :math:`K_{1,\dots,i}` is the number of unique joint values,
    and :math:`N` is the number of samples.


    Attributes
    ----------
    *data : array-like, shape (n_samples,)
        The data used to estimate the mutual information.
        You can pass an arbitrary number of data arrays as positional arguments.
    offset : int, optional
        Number of positions to shift the data arrays relative to each other.
        Delay/lag/shift between the variables. Default is no shift.
    """

    def _calculate(self):
        """Calculate the Miller-Madow mutual information of the data.

        Returns
        -------
        float
            The calculated mutual information.
        """
        return mutual_information_global(
            *self.data, log_func=self._log_base, miller_madow_correction=self.base
        )

    def _extract_local_values(self) -> ndarray[float]:
        """Separately calculate the local values.

        Returns
        -------
        ndarray[float]
            The calculated local values of mi.
        """
        return mutual_information_local(
            *self.data, log_func=self._log_base, miller_madow_correction=self.base
        )


class MillerMadowCMIEstimator(
    BaseMillerMadowMIEstimator, ConditionalMutualInformationEstimator
):
    r"""Estimator for the discrete Miller-Madow conditional mutual information.

    .. math::

        \begin{align}
        \hat{I}_{\tiny{MM}}(X_1; X_2; \ldots; X_n \mid Z)&=
        \hat{I}_{\tiny{MLE}}(\dots)
        + \frac{\left(\sum_{i=1}^{n}K_{iZ}-1\right) - \left(K_{1,\dots,i,Z}-1\right)
         - \left(K_{Z}-1\right)}{2N \cdot \log(\texttt{base})}
        \end{align}

    :math:`\hat{I}_{\tiny{MLE}}(X_1; \dots; X_n \mid Z)` is the
    initial :class:`~infomeasure.estimators.mutual_information.discrete.DiscreteCMIEstimator` estimate,
    :math:`K_{iZ}` is the number of unique values in the i-th variable joint with Z,
    :math:`K_{1,\dots,i,Z}` is the number of unique joint values,
    :math:`K_Z` is the number of unique values in the Z variable,
    and :math:`N` is the number of samples.

    Attributes
    ----------
    *data : array-like, shape (n_samples,)
        The data used to estimate the conditional mutual information.
        You can pass an arbitrary number of data arrays as positional arguments.
    cond : array-like
        The conditional data used to estimate the conditional mutual information.
    """

    def _calculate(self):
        """Calculate the conditional mutual information of the data.

        Returns
        -------
        float
            The calculated conditional mutual information.
        """
        return conditional_mutual_information_global(
            *self.data,
            cond=self.cond,
            log_func=self._log_base,
            miller_madow_correction=self.base,
        )

    def _extract_local_values(self) -> ndarray:
        """Separately calculate the local values.

        Returns
        -------
        ndarray[float]
            The calculated local values of cmi.
        """
        return conditional_mutual_information_local(
            *self.data,
            cond=self.cond,
            log_func=self._log_base,
            miller_madow_correction=self.base,
        )



"""Explicit kernel mutual information estimator tests."""

import pytest
from numpy import ndarray, equal, allclose

from tests.conftest import (
    generate_autoregressive_series,
    generate_autoregressive_series_condition,
)
from infomeasure.estimators.mutual_information import (
    KernelMIEstimator,
    KernelCMIEstimator,
)

KERNELS = ["gaussian", "box"]


@pytest.mark.parametrize("bandwidth", [0.1, 1, 10])
@pytest.mark.parametrize("kernel", KERNELS)
def test_kernel_mi(bandwidth, kernel, default_rng):
    """Test the kernel mutual information estimator."""
    data_x = default_rng.normal(0, 1, 100)
    data_y = default_rng.normal(0, 1, 100)
    est = KernelMIEstimator(data_x, data_y, bandwidth=bandwidth, kernel=kernel)
    assert isinstance(est.result(), float)
    assert isinstance(est.local_vals(), ndarray)


@pytest.mark.parametrize(
    "rng_int,bandwidth,kernel,expected",
    [
        (5, 0.01, "gaussian", 4.469608036963662),
        (5, 0.01, "box", 9.538895559637568),
        (5, 0.1, "gaussian", 0.4206212233457012),
        (5, 0.1, "box", 7.067571502526608),
        (5, 1, "gaussian", 0.04866184805475981),
        (5, 1, "box", 1.8066997932708635),
    ],
)
def test_kernel_mi_values(rng_int, bandwidth, kernel, expected):
    """Test the kernel mutual information estimator with specific values."""
    data_x, data_y = generate_autoregressive_series(rng_int, 0.5, 0.6, 0.4)
    est = KernelMIEstimator(data_x, data_y, bandwidth=bandwidth, kernel=kernel, base=2)
    assert isinstance(est.result(), float)
    assert est.result() == pytest.approx(expected)
    assert isinstance(est.local_vals(), ndarray)


@pytest.mark.parametrize(
    "rng_int,bandwidth,kernel,expected",
    [
        (5, 0.01, "gaussian", 2.7659996851),
        (5, 0.01, "box", 0.2563985),
        (5, 0.1, "gaussian", 1.76267892),
        (5, 0.1, "box", 1.7303813472),
        (5, 1, "gaussian", 0.0592152641),
        (5, 1, "box", 3.6203671293),
    ],
)
def test_kernel_cmi_values(rng_int, bandwidth, kernel, expected):
    """Test the kernel conditional mutual information estimator with specific values."""
    data_x, data_y, cond = generate_autoregressive_series_condition(
        rng_int, alpha=(0.5, 0.1), beta=0.6, gamma=(0.4, 0.2)
    )
    est = KernelCMIEstimator(
        data_x, data_y, cond=cond, bandwidth=bandwidth, kernel=kernel, base=2
    )
    assert isinstance(est.result(), float)
    assert est.result() == pytest.approx(expected)
    assert isinstance(est.local_vals(), ndarray)


@pytest.mark.parametrize("rng_int", [1, 2])
@pytest.mark.parametrize("workers", [-1, 1])
@pytest.mark.parametrize("kernel", ["gaussian", "box"])
def test_kernel_mi_parallelization(rng_int, workers, kernel):
    """Test the Kernel MI estimator with different worker counts."""
    data_x, data_y = generate_autoregressive_series(
        rng_int, 0.5, 0.6, 0.4, length=int(20001)
    )
    est_parallel = KernelMIEstimator(
        data_x,
        data_y,
        bandwidth=0.5,
        kernel=kernel,
        base=2,
        workers=workers,
    )
    est_serial = KernelMIEstimator(
        data_x,
        data_y,
        cond=None,
        bandwidth=0.5,
        kernel=kernel,
        base=2,
        workers=1,
    )
    assert est_parallel.global_val() == pytest.approx(est_serial.global_val())
    assert allclose(est_parallel.local_vals(), est_serial.local_vals())


@pytest.mark.parametrize(
    "rng_int,method,p_mi,m_cmi",
    [
        (1, "permutation_test", 0.58, 0.14),
        (1, "bootstrap", 0.02, 0.08),
        (2, "permutation_test", 0.5, 0.28),
        (2, "bootstrap", 0.04, 0.2),
        (3, "permutation_test", 1.0, 0.14),
        (4, "permutation_test", 0.98, 0.88),
    ],
)
def test_kernel_mi_statistical_test(rng_int, method, p_mi, m_cmi):
    """Test the kernel MI for p-values. Fix rng."""
    data_x, data_y, cond = generate_autoregressive_series_condition(
        rng_int, alpha=(0.5, 0.1), beta=0.6, gamma=(0.4, 0.2)
    )
    est_mi = KernelMIEstimator(
        data_x, data_y, bandwidth=0.5, kernel="box", base=2, seed=8
    )
    est_cmi = KernelCMIEstimator(
        data_x, data_y, cond=cond, bandwidth=0.5, kernel="box", base=2, seed=8
    )
    test = est_mi.statistical_test(method=method, n_tests=50)
    assert test.p_value == pytest.approx(p_mi)
    test = est_cmi.statistical_test(method=method, n_tests=50)
    assert test.p_value == pytest.approx(m_cmi)



"""Test the TE slicing utility functions."""

import pytest
from numpy import arange, array, hstack, array_equal, column_stack
from numpy.random import default_rng
from numpy.testing import assert_equal

from infomeasure.estimators.utils.te_slicing import (
    te_observations,
    cte_observations,
)


@pytest.mark.parametrize("data_len", [1, 2, 10, 100, 1e4])
@pytest.mark.parametrize("src_hist_len", [1, 2, 3])
@pytest.mark.parametrize("dest_hist_len", [1, 2, 3])
def test_te_observations_old_implementation(
    data_len, src_hist_len, dest_hist_len, step_size=1
):
    """Test the shape of the TE observations data arrays.

    Compare output to old/explicit implementation.
    The old implementation did not correctly implement the ``step_size`` subsampling.
    """
    src = arange(data_len)
    dest = arange(data_len, 2 * data_len)
    if max(src_hist_len, dest_hist_len) * step_size >= data_len:
        with pytest.raises(ValueError):
            te_observations(src, dest, src_hist_len, dest_hist_len, step_size)
        return
    (
        joint_space_data,
        data_dest_past_embedded,
        marginal_1_space_data,
        marginal_2_space_data,
    ) = te_observations(
        src,
        dest,
        src_hist_len=src_hist_len,
        dest_hist_len=dest_hist_len,
        step_size=step_size,
    )

    # Old implementation
    n = len(src)
    max_delay = max(src_hist_len * step_size, dest_hist_len * step_size)
    y_future = array([dest[i + max_delay] for i in range(n - max_delay)])
    y_history = array(
        [
            dest[i - dest_hist_len * step_size : i : step_size]
            for i in range(max_delay, n)
        ]
    )
    x_history = array(
        [src[i - src_hist_len * step_size : i : step_size] for i in range(max_delay, n)]
    )
    assert array_equal(
        joint_space_data, hstack((x_history, y_history, y_future.reshape(-1, 1)))
    )
    assert array_equal(data_dest_past_embedded, y_history)
    assert array_equal(marginal_1_space_data, hstack((x_history, y_history)))
    assert array_equal(
        marginal_2_space_data, hstack((y_history, y_future.reshape(-1, 1)))
    )


@pytest.mark.parametrize(
    "source,dest,src_hist_len,dest_hist_len,step_size,expected",
    [
        (
            arange(10),
            arange(10, 20),
            3,
            2,
            2,
            (
                [
                    [0, 2, 4, 12, 14, 16],
                    [2, 4, 6, 14, 16, 18],
                ],
                [
                    [12, 14],
                    [14, 16],
                ],
                [
                    [0, 2, 4, 12, 14],
                    [2, 4, 6, 14, 16],
                ],
                [
                    [12, 14, 16],
                    [14, 16, 18],
                ],
            ),
        ),
        (
            arange(10).reshape(-1, 1),
            arange(10, 20).reshape(-1, 1),
            3,
            2,
            2,
            (
                [
                    [0, 2, 4, 12, 14, 16],
                    [2, 4, 6, 14, 16, 18],
                ],
                [
                    [12, 14],
                    [14, 16],
                ],
                [
                    [0, 2, 4, 12, 14],
                    [2, 4, 6, 14, 16],
                ],
                [
                    [12, 14, 16],
                    [14, 16, 18],
                ],
            ),
        ),
        (
            arange(10).reshape(-1, 1, 1),
            arange(10, 20).reshape(-1, 1, 1, 1),
            3,
            2,
            2,
            (
                [
                    [0, 2, 4, 12, 14, 16],
                    [2, 4, 6, 14, 16, 18],
                ],
                [
                    [12, 14],
                    [14, 16],
                ],
                [
                    [0, 2, 4, 12, 14],
                    [2, 4, 6, 14, 16],
                ],
                [
                    [12, 14, 16],
                    [14, 16, 18],
                ],
            ),
        ),
        (
            column_stack([arange(10), arange(10)]),
            column_stack([arange(10, 20), arange(10, 20)]),
            3,
            2,
            2,
            (
                [
                    [0, 0, 2, 2, 4, 4, 12, 12, 14, 14, 16, 16],
                    [2, 2, 4, 4, 6, 6, 14, 14, 16, 16, 18, 18],
                ],
                [
                    [12, 12, 14, 14],
                    [14, 14, 16, 16],
                ],
                [
                    [0, 0, 2, 2, 4, 4, 12, 12, 14, 14],
                    [2, 2, 4, 4, 6, 6, 14, 14, 16, 16],
                ],
                [
                    [12, 12, 14, 14, 16, 16],
                    [14, 14, 16, 16, 18, 18],
                ],
            ),
        ),
        (
            column_stack([arange(10), arange(10, 20)]),
            column_stack([arange(20, 30), arange(30, 40)]),
            3,
            2,
            2,
            (
                [
                    [0, 10, 2, 12, 4, 14, 22, 32, 24, 34, 26, 36],
                    [2, 12, 4, 14, 6, 16, 24, 34, 26, 36, 28, 38],
                ],
                [
                    [22, 32, 24, 34],
                    [24, 34, 26, 36],
                ],
                [
                    [0, 10, 2, 12, 4, 14, 22, 32, 24, 34],
                    [2, 12, 4, 14, 6, 16, 24, 34, 26, 36],
                ],
                [
                    [22, 32, 24, 34, 26, 36],
                    [24, 34, 26, 36, 28, 38],
                ],
            ),
        ),
        (  # mixed dimensions
            column_stack([arange(10), arange(10, 20)]),
            arange(20, 30),
            3,
            2,
            2,
            (
                [
                    [0, 10, 2, 12, 4, 14, 22, 24, 26],
                    [2, 12, 4, 14, 6, 16, 24, 26, 28],
                ],
                [
                    [22, 24],
                    [24, 26],
                ],
                [
                    [0, 10, 2, 12, 4, 14, 22, 24],
                    [2, 12, 4, 14, 6, 16, 24, 26],
                ],
                [
                    [22, 24, 26],
                    [24, 26, 28],
                ],
            ),
        ),
        (
            arange(10),
            column_stack([arange(20, 30), arange(30, 40)]),
            3,
            2,
            2,
            (
                [
                    [0, 2, 4, 22, 32, 24, 34, 26, 36],
                    [2, 4, 6, 24, 34, 26, 36, 28, 38],
                ],
                [
                    [22, 32, 24, 34],
                    [24, 34, 26, 36],
                ],
                [
                    [0, 2, 4, 22, 32, 24, 34],
                    [2, 4, 6, 24, 34, 26, 36],
                ],
                [
                    [22, 32, 24, 34, 26, 36],
                    [24, 34, 26, 36, 28, 38],
                ],
            ),
        ),
        (
            arange(10),
            column_stack([arange(20, 30), arange(30, 40)]),
            1,
            2,
            3,
            (
                [
                    [3, 20, 30, 23, 33, 26, 36],
                    [6, 23, 33, 26, 36, 29, 39],
                ],
                [
                    [20, 30, 23, 33],
                    [23, 33, 26, 36],
                ],
                [
                    [3, 20, 30, 23, 33],
                    [6, 23, 33, 26, 36],
                ],
                [
                    [20, 30, 23, 33, 26, 36],
                    [23, 33, 26, 36, 29, 39],
                ],
            ),
        ),
        (
            column_stack([arange(20, 30), arange(30, 40)]),
            arange(10),
            2,
            1,
            1,
            (
                [
                    [20, 30, 21, 31, 1, 2],
                    [21, 31, 22, 32, 2, 3],
                    [22, 32, 23, 33, 3, 4],
                    [23, 33, 24, 34, 4, 5],
                    [24, 34, 25, 35, 5, 6],
                    [25, 35, 26, 36, 6, 7],
                    [26, 36, 27, 37, 7, 8],
                    [27, 37, 28, 38, 8, 9],
                ],
                [[1], [2], [3], [4], [5], [6], [7], [8]],
                [
                    [20, 30, 21, 31, 1],
                    [21, 31, 22, 32, 2],
                    [22, 32, 23, 33, 3],
                    [23, 33, 24, 34, 4],
                    [24, 34, 25, 35, 5],
                    [25, 35, 26, 36, 6],
                    [26, 36, 27, 37, 7],
                    [27, 37, 28, 38, 8],
                ],
                [[1, 2], [2, 3], [3, 4], [4, 5], [5, 6], [6, 7], [7, 8], [8, 9]],
            ),
        ),
    ],
)
def test_te_observations_explicit(
    source, dest, src_hist_len, dest_hist_len, step_size, expected
):
    """Test the TE observations data arrays explicitly."""
    (
        joint_space_data,
        data_dest_past_embedded,
        marginal_1_space_data,
        marginal_2_space_data,
    ) = te_observations(source, dest, src_hist_len, dest_hist_len, step_size)
    assert array_equal(joint_space_data, expected[0])
    assert array_equal(data_dest_past_embedded, expected[1])
    assert array_equal(marginal_1_space_data, expected[2])
    assert array_equal(marginal_2_space_data, expected[3])


@pytest.mark.parametrize(
    "source,dest,src_hist_len,dest_hist_len,step_size,expected",
    [
        (
            arange(5),
            arange(10, 15),
            1,
            1,
            1,
            (
                [[0], [1], [2], [3]],  # x
                [[10], [11], [12], [13]],  # y
                [[11], [12], [13], [14]],  # y'
            ),
        ),
        (
            arange(5),
            arange(10, 15),
            1,
            1,
            1,
            ([[0], [1], [2], [3]], [[10], [11], [12], [13]], [[11], [12], [13], [14]]),
        ),
        (
            arange(5),
            column_stack([arange(10, 15), arange(15, 20)]),
            1,
            1,
            2,
            (
                [[0], [2]],
                [[10, 15], [12, 17]],
                [[12, 17], [14, 19]],
            ),
        ),
        (
            arange(5),
            column_stack([arange(10, 15), arange(15, 20)]),
            1,
            2,
            2,
            ([[2]], [[10, 15, 12, 17]], [[14, 19]]),
        ),
        (
            arange(5),
            column_stack([arange(10, 15), arange(15, 20)]),
            1,
            1,
            2,
            ([[0], [2]], [[10, 15], [12, 17]], [[12, 17], [14, 19]]),
        ),
        (
            arange(20).reshape(-1, 1),
            arange(20, 40),
            3,
            2,
            3,
            (
                [[0, 3, 6], [3, 6, 9], [6, 9, 12], [9, 12, 15]],
                [[23, 26], [26, 29], [29, 32], [32, 35]],
                [[29], [32], [35], [38]],
            ),
        ),
    ],
)
def test_te_observations_explicit_separate(
    source, dest, src_hist_len, dest_hist_len, step_size, expected
):
    """Test the TE observations data arrays with spaces not constructed"""
    (
        src_history,
        dest_history,
        dest_future,
    ) = te_observations(
        source,
        dest,
        src_hist_len,
        dest_hist_len,
        step_size,
        construct_joint_spaces=False,
    )
    assert array_equal(src_history, expected[0])
    assert array_equal(dest_history, expected[1])
    assert array_equal(dest_future, expected[2])


@pytest.mark.parametrize(
    "source,dest,cond,src_hist_len,dest_hist_len,cond_hist_len,step_size,expected",
    [
        (
            arange(5),
            arange(10, 15),
            arange(20, 25),
            1,
            1,
            1,
            1,
            (
                [  # x,  y, y',  z
                    [0, 10, 11, 20],
                    [1, 11, 12, 21],
                    [2, 12, 13, 22],
                    [3, 13, 14, 23],
                ],
                [  #  y,  z
                    [10, 20],
                    [11, 21],
                    [12, 22],
                    [13, 23],
                ],
                [  # x,  y,  z
                    [0, 10, 20],
                    [1, 11, 21],
                    [2, 12, 22],
                    [3, 13, 23],
                ],
                [  #  y, y',  z
                    [10, 11, 20],
                    [11, 12, 21],
                    [12, 13, 22],
                    [13, 14, 23],
                ],
            ),
        ),
        (
            arange(5),
            arange(10, 15),
            arange(20, 25),
            1,
            1,
            2,
            1,
            (
                [  # x,  y, y', z1, z2
                    [1, 11, 12, 20, 21],
                    [2, 12, 13, 21, 22],
                    [3, 13, 14, 22, 23],
                ],
                [  #  y, z1, z2
                    [11, 20, 21],
                    [12, 21, 22],
                    [13, 22, 23],
                ],
                [  # x,  y, z1, z2
                    [1, 11, 20, 21],
                    [2, 12, 21, 22],
                    [3, 13, 22, 23],
                ],
                [  #  y, y', z1, z2
                    [11, 12, 20, 21],
                    [12, 13, 21, 22],
                    [13, 14, 22, 23],
                ],
            ),
        ),
        (
            arange(5),
            column_stack([arange(10, 15), arange(15, 20)]),
            arange(20, 25),
            1,
            1,
            1,
            2,
            (
                [  # x, ( y ) , ( y') ,  z
                    [0, 10, 15, 12, 17, 20],
                    [2, 12, 17, 14, 19, 22],
                ],
                [  # ( y ) ,  z
                    [10, 15, 20],
                    [12, 17, 22],
                ],
                [  # x, ( y ) ,  z
                    [0, 10, 15, 20],
                    [2, 12, 17, 22],
                ],
                [  # ( y ) , ( y') ,  z
                    [10, 15, 12, 17, 20],
                    [12, 17, 14, 19, 22],
                ],
            ),
        ),
        (
            arange(5),
            column_stack([arange(10, 15), arange(15, 20)]),
            arange(20, 25),
            1,
            2,
            1,
            2,
            (
                [  # x, ( y1 ), ( y2 ), ( y') ,  z
                    [2, 10, 15, 12, 17, 14, 19, 22],
                ],
                [  # ( y1 ), ( y2 ),  z
                    [10, 15, 12, 17, 22],
                ],
                [  # x, ( y1 ), ( y2 ),  z
                    [2, 10, 15, 12, 17, 22],
                ],
                [  # ( y1 ), ( y2 ), ( y') ,  z
                    [10, 15, 12, 17, 14, 19, 22],
                ],
            ),
        ),
        (
            arange(5),
            column_stack([arange(10, 15), arange(15, 20)]),
            arange(20, 25),
            1,
            1,
            2,
            2,
            (
                [  # x, ( y ) , ( y') , z1, z2
                    [2, 12, 17, 14, 19, 20, 22],
                ],
                [  # ( y ) , z1, z2
                    [12, 17, 20, 22],
                ],
                [  # x, ( y ) , z1, z2
                    [2, 12, 17, 20, 22],
                ],
                [  # ( y ) , ( y') , z1, z2
                    [12, 17, 14, 19, 20, 22],
                ],
            ),
        ),
        (
            arange(20).reshape(-1, 1),
            arange(20, 40),
            column_stack([arange(40, 60), arange(60, 80), arange(80, 100)]),
            3,
            2,
            1,
            3,
            (
                [
                    [0, 3, 6, 23, 26, 29, 46, 66, 86],
                    [3, 6, 9, 26, 29, 32, 49, 69, 89],
                    [6, 9, 12, 29, 32, 35, 52, 72, 92],
                    [9, 12, 15, 32, 35, 38, 55, 75, 95],
                ],
                [
                    [23, 26, 46, 66, 86],
                    [26, 29, 49, 69, 89],
                    [29, 32, 52, 72, 92],
                    [32, 35, 55, 75, 95],
                ],
                [
                    [0, 3, 6, 23, 26, 46, 66, 86],
                    [3, 6, 9, 26, 29, 49, 69, 89],
                    [6, 9, 12, 29, 32, 52, 72, 92],
                    [9, 12, 15, 32, 35, 55, 75, 95],
                ],
                [
                    [23, 26, 29, 46, 66, 86],
                    [26, 29, 32, 49, 69, 89],
                    [29, 32, 35, 52, 72, 92],
                    [32, 35, 38, 55, 75, 95],
                ],
            ),
        ),
    ],
)
def test_cte_observations_explicit(
    source, dest, cond, src_hist_len, dest_hist_len, cond_hist_len, step_size, expected
):
    """Test the CTE observations data arrays explicitly."""
    (
        joint_space_data,
        data_dest_past_embedded,
        marginal_1_space_data,
        marginal_2_space_data,
    ) = cte_observations(
        source, dest, cond, src_hist_len, dest_hist_len, cond_hist_len, step_size
    )
    assert array_equal(joint_space_data, expected[0])
    assert array_equal(data_dest_past_embedded, expected[1])
    assert array_equal(marginal_1_space_data, expected[2])
    assert array_equal(marginal_2_space_data, expected[3])


@pytest.mark.parametrize(
    "source,dest,cond,src_hist_len,dest_hist_len,cond_hist_len,step_size,expected",
    [
        (
            arange(5),
            arange(10, 15),
            arange(20, 25),
            1,
            1,
            1,
            1,
            (
                [[0], [1], [2], [3]],  # x
                [[10], [11], [12], [13]],  # y
                [[11], [12], [13], [14]],  # y'
                [[20], [21], [22], [23]],  # z
            ),
        ),
        (
            arange(5),
            arange(10, 15),
            arange(20, 25),
            1,
            1,
            2,
            1,
            (
                [[1], [2], [3]],
                [[11], [12], [13]],
                [[12], [13], [14]],
                [[20, 21], [21, 22], [22, 23]],
            ),
        ),
        (
            arange(5),
            column_stack([arange(10, 15), arange(15, 20)]),
            arange(20, 25),
            1,
            1,
            1,
            2,
            ([[0], [2]], [[10, 15], [12, 17]], [[12, 17], [14, 19]], [[20], [22]]),
        ),
        (
            arange(5),
            column_stack([arange(10, 15), arange(15, 20)]),
            arange(20, 25),
            1,
            2,
            1,
            2,
            ([[2]], [[10, 15, 12, 17]], [[14, 19]], [[22]]),
        ),
        (
            arange(5),
            column_stack([arange(10, 15), arange(15, 20)]),
            arange(20, 25),
            1,
            1,
            2,
            2,
            ([[2]], [[12, 17]], [[14, 19]], [[20, 22]]),
        ),
        (
            arange(20).reshape(-1, 1),
            arange(20, 40),
            column_stack([arange(40, 60), arange(60, 80), arange(80, 100)]),
            3,
            2,
            1,
            3,
            (
                [[0, 3, 6], [3, 6, 9], [6, 9, 12], [9, 12, 15]],
                [[23, 26], [26, 29], [29, 32], [32, 35]],
                [[29], [32], [35], [38]],
                [[46, 66, 86], [49, 69, 89], [52, 72, 92], [55, 75, 95]],
            ),
        ),
    ],
)
def test_cte_observations_explicit_separate(
    source, dest, cond, src_hist_len, dest_hist_len, cond_hist_len, step_size, expected
):
    """Test the CTE observations data arrays with spaces not constructed"""
    (
        src_history,
        dest_history,
        dest_future,
        cond_history,
    ) = cte_observations(
        source,
        dest,
        cond,
        src_hist_len=src_hist_len,
        dest_hist_len=dest_hist_len,
        cond_hist_len=cond_hist_len,
        step_size=step_size,
        construct_joint_spaces=False,
    )
    assert array_equal(src_history, expected[0])
    assert array_equal(dest_history, expected[1])
    assert array_equal(dest_future, expected[2])
    assert array_equal(cond_history, expected[3])


@pytest.mark.parametrize(
    "replace_idx,value,expected_error",
    [
        (3, 0, ValueError),
        (3, None, ValueError),
        (3, 1.0, ValueError),
        (4, 0, ValueError),
        (4, -3, ValueError),
        (5, 0, ValueError),
        (5, "0", ValueError),
        (6, 0, ValueError),
        (0, 1, TypeError),
        (1, [1], TypeError),
        (1, 1, TypeError),
        (2, None, TypeError),
    ],
)
def test_cte_observations_invalid_inputs(replace_idx, value, expected_error):
    """Test the CTE observations data arrays with invalid inputs."""
    arr = arange(10)
    args = [arr, arr, arr, 1, 1, 1, 1]
    args[replace_idx] = value
    with pytest.raises(expected_error):
        cte_observations(*tuple(args))


def test_te_observations_chars():
    """Test the TE observations data arrays with char arrays."""
    source = array(["a", "b", "c", "d"])
    destination = array(["e", "f", "g", "h"])
    (
        joint_space_data,
        data_dest_past_embedded,
        marginal_1_space_data,
        marginal_2_space_data,
    ) = te_observations(
        source, destination, src_hist_len=1, dest_hist_len=2, step_size=1
    )
    assert (
        array(
            [
                ["b", "e", "f", "g"],
                ["c", "f", "g", "h"],
            ]
        )
        == joint_space_data
    ).all()


def test_te_observations_tuple():
    """Test the TE observations data arrays with tuple arrays."""
    source = array([(1, 1), (2, 2), (3, 3), (4, 4)])
    destination = array([(5, 5), (6, 6), (7, 7), (8, 8)])
    (
        joint_space_data,
        data_dest_past_embedded,
        marginal_1_space_data,
        marginal_2_space_data,
    ) = te_observations(
        source, destination, src_hist_len=1, dest_hist_len=2, step_size=1
    )
    assert (
        array(
            [
                [2, 2, 5, 5, 6, 6, 7, 7],
                [3, 3, 6, 6, 7, 7, 8, 8],
            ]
        )
        == joint_space_data
    ).all()


@pytest.mark.parametrize(
    "data_len, src_hist_len, dest_hist_len, step_size",
    [
        (0, 1, 1, 1),
        (10, 0, 1, 1),
        (10, 1, 0, 1),
        (10, 1, 1, 0),
        (10, -1, 1, 1),
        (10, 1, -1, 1),
        (10, 1, 1, -1),
        (10, 1.0, 1, 1),
        (10, 1, 1.0, 1),
        (10, 1, 1, 1.0),
        (10, "1", 1, 1),
    ],
)
def test_te_observations_invalid_inputs(
    data_len, src_hist_len, dest_hist_len, step_size
):
    """Test the TE observations data arrays with invalid inputs."""
    with pytest.raises(ValueError):
        te_observations(
            arange(data_len), arange(data_len), src_hist_len, dest_hist_len, step_size
        )


@pytest.mark.parametrize(
    "src_hist_len, dest_hist_len, step_size",
    [(1, 1, 1), (2, 2, 2), (1, 4, 1), (4, 2, 1)],
)
@pytest.mark.parametrize("value", [True, default_rng(5378)])
@pytest.mark.parametrize("argument", ["permute_src", "resample_src"])
def test_te_observations_permute_resample_src(
    src_hist_len, dest_hist_len, step_size, value, argument
):
    """Test the TE observations data arrays with permutation and resampling."""
    source = arange(100)
    destination = arange(100, 200)
    (
        joint_space_data,
        data_dest_past_embedded,
        marginal_1_space_data,
        marginal_2_space_data,
    ) = te_observations(
        source,
        destination,
        src_hist_len=src_hist_len,
        dest_hist_len=dest_hist_len,
        step_size=step_size,
        permute_src=False,
    )
    (
        joint_space_data_permuted,
        data_dest_past_embedded_permuted,
        marginal_1_space_data_permuted,
        marginal_2_space_data_permuted,
    ) = te_observations(
        source,
        destination,
        src_hist_len=src_hist_len,
        dest_hist_len=dest_hist_len,
        step_size=step_size,
        **{argument: value},  # permute_src or resample_src with True or Generator
    )
    assert array_equal(  # y_i^{(k)}, \hat{y}_{i+1} fixed
        joint_space_data[:, src_hist_len + 1 :],
        joint_space_data_permuted[:, src_hist_len + 1 :],
    )
    assert not array_equal(  # permuted x_i^{(l)}
        joint_space_data[:, :src_hist_len],
        joint_space_data_permuted[:, :src_hist_len],
    )
    assert array_equal(  # y_i^{(k)} fixed
        data_dest_past_embedded, data_dest_past_embedded_permuted
    )
    assert array_equal(  # y_i^{(k)} fixed
        marginal_1_space_data[:, src_hist_len:],
        marginal_1_space_data_permuted[:, src_hist_len:],
    )
    assert not array_equal(  # permuted x_i^{(l)}
        marginal_1_space_data[:, :src_hist_len],
        marginal_1_space_data_permuted[:, :src_hist_len],
    )
    assert array_equal(  # x_i^{(k)}, \hat{y}_{i+1} fixed
        marginal_2_space_data, marginal_2_space_data_permuted
    )


def test_te_observations_permute_resample_exclusivity():
    """Test TE observations error when setting both
    ``permute_src`` and ``resample_src``."""
    with pytest.raises(ValueError):
        te_observations(
            arange(10),
            arange(10, 20),
            src_hist_len=1,
            dest_hist_len=1,
            step_size=1,
            permute_src=True,
            resample_src=True,
        )


@pytest.mark.parametrize(
    "array_len,step_size,match_str",
    [
        (3, 1, "The history demanded"),
        (10, 1.0, "must be positive integers"),  # wrong type
        (10, 0, "must be positive integers"),  # non-positive
        (10, -1, "must be positive integers"),  # non-positive
    ],
)
def test_te_observations_value_errors(array_len, step_size, match_str):
    """Test the TE observations data arrays for value errors.

    - The demanded history is greater than the length of the data.
    - Both ``step_size_src`` or ``step_size_dest`` are set along with ``step_size``.
    - They are not positive integers.
    """
    with pytest.raises(ValueError, match=match_str):
        te_observations(
            arange(array_len),
            arange(array_len),
            src_hist_len=3,
            dest_hist_len=2,
            step_size=step_size,
        )


@pytest.mark.parametrize(  # old
    "src_hist_len, dest_hist_len, step_size, expected",
    [
        (1, 1, 1, array([0, 10, 11]) + arange(9)[:, None]),
        (1, 1, 2, array([0, 10, 12]) + arange(0, 8, 2)[:, None]),
        (1, 1, 3, array([0, 10, 13]) + arange(0, 7, 3)[:, None]),
        (2, 1, 1, array([0, 1, 11, 12]) + arange(8)[:, None]),
        (1, 2, 1, array([1, 10, 11, 12]) + arange(8)[:, None]),
        (1, 1, 2, array([0, 10, 12]) + arange(0, 8, 2)[:, None]),
        (2, 1, 2, array([0, 2, 12, 14]) + arange(0, 6, 2)[:, None]),
        (3, 1, 2, array([0, 2, 4, 14, 16]) + arange(0, 4, 2)[:, None]),
        (2, 2, 2, array([0, 2, 10, 12, 14]) + arange(0, 6, 2)[:, None]),
        (1, 2, 2, array([2, 10, 12, 14]) + arange(0, 6, 2)[:, None]),
        (3, 2, 2, array([0, 2, 4, 12, 14, 16]) + arange(0, 4, 2)[:, None]),
        (3, 2, 3, array([0, 3, 6, 13, 16, 19]) + arange(0, 1, 2)[:, None]),
    ],
)
def test_te_observations_step_sizes(src_hist_len, dest_hist_len, step_size, expected):
    """Test the TE observations data arrays with different step sizes."""
    joint_space_data, _, _, _ = te_observations(
        arange(10),
        arange(10, 20),
        src_hist_len=src_hist_len,
        dest_hist_len=dest_hist_len,
        step_size=step_size,
    )
    assert_equal(joint_space_data, expected)



"""Module for the kernel entropy estimator."""

from numpy import column_stack, sum as np_sum, isnan, nan

from ... import Config
from ...utils.types import LogBaseType
from ..base import EntropyEstimator
from ..mixins import WorkersMixin
from ..utils.array import assure_2d_data
from ..utils.kde import kde_probability_density_function


class KernelEntropyEstimator(WorkersMixin, EntropyEstimator):
    r"""Kernel entropy estimator for continuous data using Kernel Density Estimation (KDE).

    The kernel entropy estimator computes the differential Shannon entropy by estimating
    the probability density function using kernel density estimation:

    .. math::

        \hat{H}(X) = -\int \hat{f}(x) \log \hat{f}(x) \, dx \approx -\frac{1}{N} \sum_{i=1}^{N} \log \hat{f}(x_i)

    where :math:`\hat{f}(x)` is the kernel density estimate:

    .. math::

        \hat{f}(x) = \frac{1}{N h^d} \sum_{i=1}^{N} K\left(\frac{x - x_i}{h}\right)

    with :math:`K(\cdot)` being the kernel function, :math:`h` the bandwidth parameter,
    :math:`d` the dimensionality, and :math:`N` the number of data points.

    For joint entropy of multiple variables, the estimator concatenates the variables
    into a single multivariate space and applies the same KDE approach.

    The estimator supports both Gaussian and box (uniform) kernels. The choice of
    bandwidth is critical: small values can lead to under-smoothing and overfitting,
    while large values may over-smooth the data and obscure important features
    :cite:p:`silverman1986density,garcia-portuguesChapter2Kernel2025`.

    Parameters
    ----------
    *data : array-like
        The continuous data used to estimate the entropy. For univariate entropy,
        pass a single array. For joint entropy, pass multiple arrays.
    bandwidth : float | int
        The bandwidth parameter for the kernel. Controls the smoothness of the
        density estimate.
    kernel : str
        Type of kernel to use. Supported options are:

        - ``'gaussian'``: Gaussian (normal) kernel
        - ``'box'``: Box (uniform) kernel

        Compatible with the KDE implementation
        :func:`kde_probability_density_function() <infomeasure.estimators.utils.kde.kde_probability_density_function>`.
    workers : int, optional
        Number of workers to use for parallel processing. Default is 1 (no parallelization).
        If set to -1, all available CPU cores will be used.
    base : float | str, optional
        Logarithm base for entropy calculation. Default is from global configuration.

    Attributes
    ----------
    *data : array-like
        The data used to estimate the entropy.
    bandwidth : float | int
        The bandwidth for the kernel.
    kernel : str
        Type of kernel to use.
    workers : int
        Number of workers to use for parallel processing.

    Returns
    -------
    array-like
        Local entropy values for each data point when calling entropy calculation methods.
        The mean of these values gives the overall entropy estimate.

    Notes
    -----
    **Bandwidth Selection**: The bandwidth parameter critically affects the quality of
    the entropy estimate. A small bandwidth can lead to under-sampling and high variance,
    while a large bandwidth may over-smooth the data, obscuring important details and
    introducing bias.

    **Kernel Choice**:

    - Gaussian kernels provide smooth density estimates and are theoretically well-founded
    - Box kernels are computationally efficient and provide non-parametric estimates

    **Computational Complexity**: The algorithm has O(N²) complexity for box kernels
    using KDTree queries, and varies for Gaussian kernels depending on the implementation.

    **Cross-entropy**: Supported between two distributions by evaluating the density
    of the second distribution at points from the first distribution.

    Examples
    --------
    >>> import infomeasure as im
    >>> from numpy.random import default_rng
    >>> rng = default_rng(281769)
    >>> # Generate sample data
    >>> data = rng.normal(0, 1, 1000)
    >>>
    >>> # Create estimator
    >>> estimator = im.estimator(data, measure="h", approach="kernel", bandwidth=0.5, kernel='gaussian')
    >>>
    >>> # Calculate entropy
    >>> estimator.result()
    np.float64(1.366015332652949)
    >>> # Local values
    >>> estimator.local_vals()
    array([1.54017083, 1.35855839, 0.97949819, 0.97333173, 2.62084886,
       ...
       1.08174049, 0.97418054, 1.88055967, 0.99614516, 0.98548583])


    See Also
    --------
    infomeasure.estimators.utils.kde.kde_probability_density_function :
        Underlying KDE implementation
    infomeasure.estimators.entropy.discrete.DiscreteEntropyEstimator :
        For discrete data entropy estimation
    """

    def __init__(
        self,
        *data,
        bandwidth: float | int,
        kernel: str,
        workers: int = 1,
        base: LogBaseType = Config.get("base"),
    ):
        """Initialize the KernelEntropyEstimator.

        Parameters
        ----------
        bandwidth : float | int
            The bandwidth for the kernel.
        kernel : str
            Type of kernel to use, compatible with the KDE
            implementation :func:`kde_probability_density_function() <infomeasure.estimators.utils.kde.kde_probability_density_function>`.
        workers : int, optional
           Number of workers to use for parallel processing.
           Default is 1, meaning no parallel processing.
           If set to -1, all available CPU cores will be used.
        """
        super().__init__(*data, workers=workers, base=base)
        self.data = tuple(assure_2d_data(var) for var in self.data)
        self.bandwidth = bandwidth
        self.kernel = kernel

    def _simple_entropy(self):
        """Calculate the entropy of the data.

        Returns
        -------
        array-like
            The local form of the entropy.
        """
        # Compute the KDE densities
        densities = kde_probability_density_function(
            self.data[0], self.bandwidth, kernel=self.kernel, workers=self.n_workers
        )
        densities[densities == 0] = nan
        # Compute the log of the densities
        log_densities = -self._log_base(densities)
        log_densities[isnan(log_densities)] = 0
        return log_densities

    def _joint_entropy(self):
        """Calculate the joint entropy of the data.

        This is done by joining the variables into one space
        and calculating the entropy.

        Returns
        -------
        array-like
            The local form of the joint entropy.
        """
        self.data = (column_stack(self.data[0]),)
        return self._simple_entropy()

    def _cross_entropy(self) -> float:
        """Calculate the cross-entropy between two distributions.

        Returns
        -------
        float
            The calculated cross-entropy.
        """
        # Compute the KDE densities
        densities = kde_probability_density_function(
            self.data[1],
            self.bandwidth,
            at=self.data[0],
            kernel=self.kernel,
            workers=self.n_workers,
        )
        # Compute the log of the densities
        return -np_sum(self._log_base(densities[densities > 0])) / len(densities)



"""Functions for interaction information, a multivariate generalization of
mutual information."""

from collections import Counter

from numpy import (
    clip,
    uint64,
    log,
    ndarray,
    ones,
    prod,
    ravel,
    unique,
    zeros,
    count_nonzero,
)
from numpy import (
    sum as np_sum,
)
from scipy.sparse import find as sp_find
from scipy.stats.contingency import crosstab
from sparse import COO, asnumpy


def mutual_information_global(
    *data: tuple,
    log_func: callable = log,
    miller_madow_correction: str | float | int = None,
) -> float:
    """Estimate the global mutual information between multiple random variables.

    Parameters
    ----------
    *data : array-like, shape (n_samples,)
        The data used to estimate the global mutual information.
        You can pass an arbitrary number of data arrays as positional arguments.
    log_func : callable, optional
        The logarithm function to use. Default is the natural logarithm.
    miller_madow_correction : str | float | int, optional
        If not None, apply the Miller-Madow correction to the global mutual
        information in the information unit of the passed value.
        ``log_func`` and ``miller_madow_correction`` should be the same base.

    Returns
    -------
    float
        The global mutual information between the random variables.
    """
    # check that miller_madow_correction is None, float, int or "e"
    if miller_madow_correction is not None and (
        not isinstance(miller_madow_correction, (str, float, int))
        or (isinstance(miller_madow_correction, str) and miller_madow_correction != "e")
    ):
        raise ValueError(
            f"miller_madow_correction must be None, float or 'e', "
            f"got {miller_madow_correction}."
        )

    if all(d.ndim == 1 for d in data):
        if len(data) == 2:
            return _mutual_information_global_2d_int(
                *data,
                log_func=log_func,
                miller_madow_correction=miller_madow_correction,
            )
        else:
            return _mutual_information_global_nd_int(
                *data,
                log_func=log_func,
                miller_madow_correction=miller_madow_correction,
            )
    else:
        return _mutual_information_global_nd_other(
            *data,
            log_func=log_func,
            miller_madow_correction=miller_madow_correction,
        )


def _mutual_information_global_nd_int(
    *data: tuple,
    log_func: callable = log,
    miller_madow_correction: str | float | int = None,
) -> float:
    """Estimate the global mutual information between an arbitrary number of
    random variables."""
    uniques, indices = zip(*[unique(var, return_inverse=True, axis=0) for var in data])
    contingency_coo = COO(
        coords=indices,
        data=ones(len(indices[0]), dtype=uint64),
        shape=tuple(len(uniq) for uniq in uniques),
        fill_value=0,
    )

    # Non-zero indices and values
    idxs = contingency_coo.nonzero()
    vals = contingency_coo.data

    # Marginal probabilities
    count_marginals = [
        asnumpy(
            contingency_coo.sum(  # all axes, except i
                axis=tuple(range(contingency_coo.ndim))[:i]
                + tuple(range(contingency_coo.ndim))[i + 1 :]
            )
        )
        for i in range(contingency_coo.ndim)
    ]

    # Early return if any of the marginal entropies is zero
    if any(count_m.size == 1 for count_m in count_marginals):
        return 0.0

    # Calculate the expected logarithm values for the outer product of marginal
    # probabilities, only for non-zero entries.
    outer = prod(
        [
            count.take(idx).astype(uint64, copy=False)
            for count, idx in zip(count_marginals, idxs)
        ],
        axis=0,
    )

    # Normalized contingency table (joint probability)
    contingency_sum = contingency_coo.sum()
    p_joint = vals / contingency_sum

    # Logarithm of the non-zero elements
    log_p_joint = log_func(vals)

    log_outer = -log_func(outer) + len(count_marginals) * log_func(contingency_sum)
    # Combine the terms to calculate the mutual information
    mi = p_joint * (log_p_joint - log_func(contingency_sum)) + p_joint * log_outer

    misum = mi.sum()  # interaction information can be negative, do not clip
    if miller_madow_correction is None:
        return misum
    else:
        corr = millermadow_mi_corr(
            contingency_coo.shape, len(vals), len(data[0]), miller_madow_correction
        )
        return misum + corr


def _mutual_information_global_2d_int(
    *data: tuple[ndarray[int]],
    log_func: callable = log,
    miller_madow_correction: str | float | int = None,
) -> float:
    """Estimate the global mutual information between two random variables.

    The approach relies on the contingency table of the two variables.
    Instead of calculating the full outer product, only the non-zero elements are
    considered.
    Code adapted from
    the :func:`mutual_info_score() <sklearn.metrics.mutual_info_score>` function in
    scikit-learn.
    """
    # Contingency table - COOrdinate sparse matrix
    contingency_coo = crosstab(*data, sparse=True).count
    # Non-zero indices and values
    nzx, nzy, nzv = sp_find(contingency_coo)

    # Normalized contingency table (joint probability)
    contingency_sum = contingency_coo.sum()
    p_joint = nzv / contingency_sum
    # Marginal probabilities
    pi = ravel(contingency_coo.sum(axis=1))
    pj = ravel(contingency_coo.sum(axis=0))

    # Early return if any of the marginal entropies is zero
    if pi.size == 1 or pj.size == 1:
        return 0.0

    # Logarithm of the non-zero elements
    log_p_joint = log_func(nzv)

    # Calculate the expected logarithm values for the outer product of marginal
    # probabilities, only for non-zero entries.
    outer = pi.take(nzx).astype(uint64, copy=False) * pj.take(nzy).astype(
        uint64, copy=False
    )
    log_outer = -log_func(outer) + log_func(pi.sum()) + log_func(pj.sum())
    # Combine the terms to calculate the mutual information
    mi = p_joint * (log_p_joint - log_func(contingency_sum)) + p_joint * log_outer
    misum = clip(mi.sum(), 0.0, None)
    if miller_madow_correction is None:
        return misum
    else:
        corr = millermadow_mi_corr(
            contingency_coo.shape, len(nzv), len(data[0]), miller_madow_correction
        )
        return misum + corr


def _mutual_information_global_nd_other(
    *data: tuple[ndarray],
    log_func: callable = log,
    miller_madow_correction: str | float | int = None,
) -> float:
    """Alternative method to estimate the global mutual information between an
    arbitrary number of random variables.

    Same as :func:`_mutual_information_global_nd_int`, but for non-integer data.
    """
    # data is a tuple of ndarrays, for joint data, concatenate these rows
    # joint_data = [tuple(row) for row in column_stack(data)]
    joint_data = [tuple(tuple(val) for val in row) for row in zip(*data)]

    # Count joint and marginal occurrences
    joint_counts = Counter(joint_data)
    joint_total = sum(joint_counts.values())
    marginal_counts = [Counter([tuple(val) for val in var]) for var in data]
    marginal_totals = [sum(counts.values()) for counts in marginal_counts]

    # Estimate probabilities
    joint_prob = {key: val / joint_total for key, val in joint_counts.items()}
    marginal_prob = [
        {key: val / total for key, val in counts.items()}
        for counts, total in zip(marginal_counts, marginal_totals)
    ]

    # Calculate the mutual information
    mi_sum = [
        joint_prob[key]
        * log_func(
            joint_prob[key]
            / prod([marginal_prob[i][key[i]] for i in range(len(data))], axis=0)
        )
        for key in joint_prob
    ]
    if len(mi_sum) == 0:
        return 0.0
    misum = np_sum(mi_sum)

    if miller_madow_correction is None:
        return misum
    else:
        corr = millermadow_mi_corr(
            list(len(counter.keys()) for counter in marginal_counts),
            len(joint_counts),
            len(data[0]),
            miller_madow_correction,
        )
        return misum + corr


def mutual_information_local(
    *data: tuple,
    log_func: callable = log,
    miller_madow_correction: str | float | int = None,
) -> ndarray:
    """Estimate the local mutual information between multiple random variables.

    The mean of the local mutual information is the global mutual information.
    Only calculating the global value is more efficient,
    so evaluating the local mutual information should only be done
    when explicitly needed.

    Parameters
    ----------
    *data : array-like, shape (n_samples,)
        The data used to estimate the local mutual information.
        You can pass an arbitrary number of data arrays as positional arguments.
    log_func : callable, optional
        The logarithm function to use. Default is the natural logarithm.
    miller_madow_correction : str | float | int, optional
        If not None, apply the Miller-Madow correction to the global mutual
        information in the information unit of the passed value.
        ``log_func`` and ``miller_madow_correction`` should be the same base.

    Returns
    -------
    ndarray
        The local mutual information between the random variables.
    """
    # check that miller_madow_correction is None, float, int or "e"
    if miller_madow_correction is not None and (
        not isinstance(miller_madow_correction, (str, float, int))
        or (isinstance(miller_madow_correction, str) and miller_madow_correction != "e")
    ):
        raise ValueError(
            f"miller_madow_correction must be None, float or 'e', "
            f"got {miller_madow_correction}."
        )
    # Contingency table - COOrdinate sparse matrix
    uniques, indices = zip(*[unique(var, return_inverse=True, axis=0) for var in data])
    contingency_coo = COO(
        coords=indices,
        data=ones(len(indices[0]), dtype=uint64),
        shape=tuple(len(uniq) for uniq in uniques),
        fill_value=0,
    )

    # Normalized contingency table (joint probability)
    contingency_sum = contingency_coo.sum()
    # Marginal probabilities
    count_marginals = [
        asnumpy(
            contingency_coo.sum(  # all axes, except i
                axis=tuple(range(contingency_coo.ndim))[:i]
                + tuple(range(contingency_coo.ndim))[i + 1 :]
            )
        )
        for i in range(contingency_coo.ndim)
    ]

    # Early return if any of the marginal entropies is zero
    if any(count_m.size == 1 for count_m in count_marginals):
        return zeros(len(data[0]))

    # To get local values we iterate over *data
    # for each row in the input data: log( p(data) / p(data1) * p(data2) )
    p_joint = contingency_coo[indices].data / contingency_sum
    outer = prod(
        [count[indices[i]] for i, count in enumerate(count_marginals)]
        / contingency_sum,
        axis=0,
    )
    mi_local = -log_func(outer) + log_func(p_joint)
    if miller_madow_correction is None:
        return mi_local
    else:
        corr = millermadow_mi_corr(
            contingency_coo.shape,
            len(contingency_coo.data),
            len(data[0]),
            miller_madow_correction,
        )
        return mi_local + corr


def conditional_mutual_information_global(
    *data: tuple,
    cond: ndarray,
    log_func: callable = log,
    miller_madow_correction: str | float | int = None,
) -> float:
    """Estimate the global conditional mutual information
    between multiple random variables and a conditioning variable.

    Parameters
    ----------
    *data : array-like, shape (n_samples,)
        The data used to estimate the global mutual information.
        You can pass an arbitrary number of data arrays as positional arguments.
    cond : array-like, shape (n_samples,)
        The conditioning variable.
    log_func : callable, optional
        The logarithm function to use. Default is the natural logarithm.
    miller_madow_correction : str | float | int, optional
        If not None, apply the Miller-Madow correction to the global mutual
        information in the information unit of the passed value.
        ``log_func`` and ``miller_madow_correction`` should be the same base.

    Returns
    -------
    float
        The global conditional mutual information between the random variables.

    Raises
    ------
    ValueError
        If the conditioning variable is not one-dimensional.
    """
    # check that miller_madow_correction is None, float, int or "e"
    if miller_madow_correction is not None and (
        not isinstance(miller_madow_correction, (str, float, int))
        or (isinstance(miller_madow_correction, str) and miller_madow_correction != "e")
    ):
        raise ValueError(
            f"miller_madow_correction must be None, float or 'e', "
            f"got {miller_madow_correction}."
        )
    if cond.ndim != 1:
        raise ValueError("The conditioning variable must be one-dimensional.")
    return _conditional_mutual_information_global_nd_int(
        *data,
        cond=cond,
        log_func=log_func,
        miller_madow_correction=miller_madow_correction,
    )


def _conditional_mutual_information_global_nd_int(
    *data: tuple,
    cond: ndarray,
    log_func: callable = log,
    miller_madow_correction: str | float | int = None,
) -> float:
    """Estimate the global conditional mutual information between an arbitrary number of
    random variables and a conditioning variable."""
    uniques, indices = zip(
        *[unique(var, return_inverse=True, axis=0) for var in (data + (cond,))]
    )
    contingency_coo = COO(
        coords=indices,
        data=ones(len(indices[0]), dtype=uint64),
        shape=tuple(len(uniq) for uniq in uniques),
        fill_value=0,
    )

    # Non-zero indices and values
    idxs = contingency_coo.nonzero()
    vals = contingency_coo.data

    # Marginal-conditioned probabilities
    count_marginals_cond = [
        asnumpy(
            contingency_coo.sum(  # all axes, except i and cond
                axis=tuple(range(contingency_coo.ndim - 1))[:i]
                + tuple(range(contingency_coo.ndim - 1))[i + 1 :]
            )
        )
        for i in range(contingency_coo.ndim - 1)
    ]
    count_cond = asnumpy(
        contingency_coo.sum(axis=tuple(range(contingency_coo.ndim - 1)))
    )
    # all axes, except cond

    # Early return if any of the marginal entropies is zero
    if any(count_m.size == 1 for count_m in count_marginals_cond):
        return 0.0

    # Calculate the expected logarithm values for the outer product of marginal
    # probabilities, only for non-zero entries.
    outer = prod(
        [
            count[idx, idxs[-1]].astype(uint64, copy=False)
            for count, idx in zip(count_marginals_cond, idxs[:-1])
        ],
        axis=0,
    )

    # Normalized contingency table (joint probability)
    contingency_sum = contingency_coo.sum()
    p_joint = vals / contingency_sum

    # Logarithm of the non-zero elements
    p_cond = count_cond.take(idxs[-1]).astype(uint64, copy=False)
    log_p_joint = log_func(vals * p_cond) - log_func(contingency_sum)

    log_outer = -log_func(outer) + sum(
        log_func(count_m.sum()) for count_m in count_marginals_cond[:-1]
    )
    # Combine the terms to calculate the mutual information
    mi = p_joint * log_p_joint + p_joint * log_outer
    misum = mi.sum()  # interaction information can be negative, do not clip
    if miller_madow_correction is None:
        return misum
    else:
        corr = millermadow_mi_corr(
            list(count_nonzero(marg) for marg in count_marginals_cond),
            len(vals),
            len(data[0]),
            miller_madow_correction,
            k_cond=len(count_cond),
        )
        return misum + corr


def conditional_mutual_information_local(
    *data: tuple,
    cond: ndarray,
    log_func: callable = log,
    miller_madow_correction: str | float | int = None,
) -> ndarray:
    """Estimate the local conditional mutual information between multiple
    random variables and a conditioning variable.

    The mean of the local conditional mutual information is the
    global conditional mutual information.
    Only calculating the global value is more efficient,
    so evaluating the local conditional mutual information should only be done
    when explicitly needed.

    Parameters
    ----------
    *data : array-like, shape (n_samples,)
        The data used to estimate the local mutual information.
        You can pass an arbitrary number of data arrays as positional arguments.
    cond : array-like, shape (n_samples,)
        The conditioning variable.
    log_func : callable, optional
        The logarithm function to use. Default is the natural logarithm.
    miller_madow_correction : str | float | int, optional
        If not None, apply the Miller-Madow correction to the global mutual
        information in the information unit of the passed value.
        ``log_func`` and ``miller_madow_correction`` should be the same base.

    Returns
    -------
    ndarray
        The local conditional mutual information between the random variables.
    """
    # check that miller_madow_correction is None, float, int or "e"
    if miller_madow_correction is not None and (
        not isinstance(miller_madow_correction, (str, float, int))
        or (isinstance(miller_madow_correction, str) and miller_madow_correction != "e")
    ):
        raise ValueError(
            f"miller_madow_correction must be None, float or 'e', "
            f"got {miller_madow_correction}."
        )
    # Contingency table - COOrdinate sparse matrix
    uniques, indices = zip(
        *[unique(var, return_inverse=True, axis=0) for var in (data + (cond,))]
    )
    contingency_coo = COO(
        coords=indices,
        data=ones(len(indices[0]), dtype=uint64),
        shape=tuple(len(uniq) for uniq in uniques),
        fill_value=0,
    )

    # Normalized contingency table (joint probability)
    contingency_sum = contingency_coo.sum()
    # Marginal-conditioned probabilities
    count_marginals_cond = [
        asnumpy(
            contingency_coo.sum(  # all axes, except i and cond
                axis=tuple(range(contingency_coo.ndim - 1))[:i]
                + tuple(range(contingency_coo.ndim - 1))[i + 1 :]
            )
        )
        for i in range(contingency_coo.ndim - 1)
    ]
    count_cond = asnumpy(
        contingency_coo.sum(axis=tuple(range(contingency_coo.ndim - 1)))
    )

    # Early return if any of the marginal entropies is zero
    if any(count_m.size == 1 for count_m in count_marginals_cond):
        return zeros(len(data[0])).astype(float)

    # To get local values we iterate over *data
    # for each row in the input data: log( p(data) / p(data1) * p(data2) )
    p_joint = contingency_coo[indices].data / contingency_sum
    p_cond = count_cond[indices[-1]] / contingency_sum
    outer = prod(
        [count[indices[i], indices[-1]] for i, count in enumerate(count_marginals_cond)]
        / contingency_sum,
        axis=0,
    )
    misum = log_func(p_joint * p_cond) - log_func(outer)
    if miller_madow_correction is None:
        return misum
    else:
        corr = millermadow_mi_corr(
            list(count_nonzero(marg) for marg in count_marginals_cond),
            len(contingency_coo.data),
            len(data[0]),
            miller_madow_correction,
            k_cond=len(count_cond),
        )
        return misum + corr


def millermadow_mi_corr(k_i, k_joint, n, base, k_cond=None):
    """
    Computes the Miller-Madow mutual information correction term.

    This function calculates a correction term used to adjust the bias in mutual
    information estimates, which arise due to finite sample size issues.
    The correction is based on the marginal counts and joint count in the observed
    distributions.

    Parameters
    ----------
    k_i : list[int]
        A list containing the marginal cardinalities of individual variables in the
        dataset.
        Each element represents the number of unique values for the respective
        variable.
    k_joint : int
        The cardinality of the joint distribution.
        Represents the number of unique
        observations across all combined dimensions.
    n : int
        The sample size, representing the total number of observations in the data.
    base : str | float | int
        The logarithmic base used for the mutual information computation.
        If set to
        "e", natural logarithm is used.
        Otherwise, log of the specified base is used.
    k_cond : int, optional
        The cardinality of the conditional variable.
        When this is used, k_i can be used as k_iZ, to calculate the correction for
        conditional MI.

    Returns
    -------
    float
        The calculated Miller-Madow correction term to adjust the mutual information
        value.
    """
    corr = (
        sum(k_i)  # sum K_i
        - len(k_i)  # sum -1
        - k_joint
        + 1  # -(K_{1,...,i} +1)
        - ((k_cond - 1) if k_cond is not None else 0)
    ) / (2 * n)  # / 2N
    if base != "e":
        corr /= log(base)
    return corr



"""Explicit discrete entropy estimator tests."""

import pytest
from numpy import e, log

from infomeasure import entropy, estimator


@pytest.mark.parametrize(
    "data,base,expected",
    [
        ([1, 1, 1, 1, 1], 2, 0.0),
        ([1, 0, 1, 0], 2, 1.0),
        (["a", 0, "a", 0], 2, 1.0),
        ([[0, 3], [4, 4], [4, 4], [0, 3]], 2, 1.0),
        ([[0, 3], [4, 4], [4, 4], [3, 0]], 2, 1.5),
        ([1, 2, 3, 4, 5], 2, 2.321928094887362),
        ([1, 2, 3, 4, 5], 10, 0.6989700043360187),
        ([1, 2, 3, 4, 5], "e", 1.6094379124341003),
    ],
)
def test_discrete_entropy(data, base, expected):
    """Test the discrete entropy estimator."""
    assert entropy(data, approach="discrete", base=base) == pytest.approx(expected)


@pytest.mark.parametrize(
    "data,base,expected",
    [
        (([1, 1, 1, 1, 1], [1, 1, 1, 1, 1]), 2, 0.0),
        (([1, 1, 1, 1, 1], [1, 1, 1, 1, 1], [1, 1, 1, 1, 1]), 2, 0.0),
        (([1, 1, 7, 2, 3, 6, 6, 3], [2, 3, 6, 6, 3, 6, 5, 7]), 2, 3.0),
        (([1, 1, 7, 2, 3, 6, 6, 3], [2, 3, 6, 6, 3, 6, 5, 7]), 10, 0.90308999),
        (([1, 2, 3 - 2] * 1000, [1, 2, 3 - 2] * 1000), 10, 0.276434591),
    ],
)
def test_discrete_joint_entropy(data, base, expected):
    """Test the discrete joint entropy estimator."""
    est = estimator(data, measure="entropy", approach="discrete", base=base)
    assert est.result() == pytest.approx(expected)
    est.local_vals()


# try different bases with uniform distribution
@pytest.mark.parametrize("length", [1, 2, 10, 100, 1000])
@pytest.mark.parametrize("base", [2, 2.5, 3, 10, e])
def test_discrete_entropy_uniform(length, base):
    r"""Test the discrete entropy estimator with a uniform distribution.

    The entropy of a uniform distribution is given by:

    :math:`H(X) = -\log_b(1/n) = \log_b(n)`
    """
    data = range(0, length)
    assert entropy(data, approach="discrete", base=base) == pytest.approx(
        log(length) / log(base)
    )


@pytest.mark.parametrize(
    "data_p, data_q",
    [
        ([0, 1], [0, 1]),
        ([0, 1], [1, 0]),
        ([0, 1, 2], [0, 1, 2]),
        ([0, 1, 2], [2, 1, 0]),
    ],
)
def test_discrete_cross_entropy(data_p, data_q):
    """Test the discrete cross-entropy estimator."""
    assert entropy(data_p, data_q, approach="discrete") > 0



"""Custom exceptions for the infomeasure package."""


class TheoreticalInconsistencyError(Exception):
    """Exception raised when a method cannot be implemented due to theoretical inconsistencies.

    This exception is used for cases where the mathematical or theoretical foundation
    of a method makes it inappropriate or impossible to implement in a meaningful way.
    Examples include cross-entropy for estimators that mix bias corrections from
    different distributions, or methods that violate fundamental assumptions.

    This is different from NotImplementedError, which indicates that implementation
    is planned but not yet done. TheoreticalInconsistencyError indicates that
    implementation is not theoretically sound or appropriate.
    """

    pass



"""Test configuration class for the infomeasure module."""

import pytest

from infomeasure.utils import Config


class TestConfig:
    """Test configuration class for the infomeasure module."""

    def test_get(self):
        """Test the get method of the Config class."""
        assert Config.get("base") == "e"

    def test_set(self):
        """Test the set method of the Config class."""
        Config.set("base", 10)
        assert Config.get("base") == 10
        Config.set("base", 2)
        assert Config.get("base") == 2
        Config.set("base", "e")
        assert Config.get("base") == "e"

    def test_set_invalid_key(self):
        """Test setting an invalid key."""
        with pytest.raises(KeyError):
            Config.set("invalid_key", 10)

    def test_get_invalid_key(self):
        """Test getting an invalid key."""
        with pytest.raises(KeyError):
            Config.get("invalid_key")

    def test_set_invalid_value(self):
        """Test setting an invalid value."""
        with pytest.raises(TypeError):
            Config.set("base", "invalid_value")

    @pytest.mark.parametrize(
        "unit, expected_base",
        [
            ("bits", 2),
            ("dits", 10),
            ("shannon", 2),
            ("shannons", 2),
            ("nats", "e"),
            ("hartley", 10),
            ("nat", "e"),
            ("hartleys", 10),
            ("bans", 10),
            ("ban", 10),
            ("bit", 2),
            ("dit", 10),
        ],
    )
    def test_set_get_logarithmic_unit(self, unit, expected_base):
        """Test the set_logarithmic_unit method of the Config class."""
        Config.set_logarithmic_unit(unit)
        assert Config.get("base") == expected_base
        assert (
            unit[:-1] if unit.endswith("s") else unit
        ) in Config.get_logarithmic_unit()
        Config.get_logarithmic_unit_description()

    def test_get_unknown_unit(self):
        """Test getting an unknown logarithmic unit and description."""
        Config.set("base", 3)
        assert Config.get_logarithmic_unit() == "unknown (base 3)"
        with pytest.raises(
            ValueError, match="No description for logarithmic unit: base 3"
        ):
            Config.get_logarithmic_unit_description()

    def test_set_logarithmic_unit_invalid(self):
        """Test setting an invalid logarithmic unit."""
        with pytest.raises(ValueError):
            Config.set_logarithmic_unit("invalid_unit")

    def test_reset(self):
        """Test the reset method of the Config class."""
        Config.set("base", 10)
        Config.reset()
        assert Config.get("base") == "e"



"""Module for test fixtures available for all test files"""

from functools import cache

import pytest
from numpy import roll, zeros
from numpy.random import default_rng as rng

from infomeasure import Config
from infomeasure.estimators import entropy, mutual_information, transfer_entropy

# Dictionary for each measure with the needed kwargs for the test
# ``functional_str`` should contain all the strings that can be used to address the
#  estimator in the functional API.
# ``needed_kwargs`` should only contain the kwargs that need to be passed to the
#  estimator. All other kwargs should be tested in each dedicated estimator test file.

ENTROPY_APPROACHES = {
    "AnsbEntropyEstimator": {
        "functional_str": ["ansb"],
        "needed_kwargs": {},
    },
    "BayesEntropyEstimator": {
        "functional_str": ["bayes"],
        "needed_kwargs": {"alpha": 1.0},
    },
    "BonachelaEntropyEstimator": {
        "functional_str": ["bonachela"],
        "needed_kwargs": {},
    },
    "ChaoShenEntropyEstimator": {
        "functional_str": ["chao_shen", "cs"],
        "needed_kwargs": {},
    },
    "ChaoWangJostEntropyEstimator": {
        "functional_str": ["chao_wang_jost", "cwj"],
        "needed_kwargs": {},
    },
    "DiscreteEntropyEstimator": {
        "functional_str": ["discrete"],
        "needed_kwargs": {},
    },
    "GrassbergerEntropyEstimator": {
        "functional_str": ["grassberger"],
        "needed_kwargs": {},
    },
    "KernelEntropyEstimator": {
        "functional_str": ["kernel"],
        "needed_kwargs": {"bandwidth": 0.3, "kernel": "box"},
    },
    "KozachenkoLeonenkoEntropyEstimator": {
        "functional_str": ["metric", "kl"],
        "needed_kwargs": {},
    },
    "MillerMadowEntropyEstimator": {
        "functional_str": ["miller_madow", "mm"],
        "needed_kwargs": {},
    },
    "NsbEntropyEstimator": {
        "functional_str": ["nsb"],
        "needed_kwargs": {},
    },
    "OrdinalEntropyEstimator": {
        "functional_str": ["ordinal", "symbolic", "permutation"],
        "needed_kwargs": {"embedding_dim": 2},
    },
    "RenyiEntropyEstimator": {
        "functional_str": ["renyi"],
        "needed_kwargs": {"alpha": 1.5},
    },
    "ShrinkEntropyEstimator": {
        "functional_str": ["shrink", "js"],
        "needed_kwargs": {},
    },
    "TsallisEntropyEstimator": {
        "functional_str": ["tsallis"],
        "needed_kwargs": {"q": 2.0},
    },
    "ZhangEntropyEstimator": {
        "functional_str": ["zhang"],
        "needed_kwargs": {},
    },
}

MI_APPROACHES = {
    "AnsbMIEstimator": {
        "functional_str": ["ansb"],
        "needed_kwargs": {},
    },
    "BayesMIEstimator": {
        "functional_str": ["bayes"],
        "needed_kwargs": {"alpha": "jeffrey"},
    },
    "BonachelaMIEstimator": {
        "functional_str": ["bonachela"],
        "needed_kwargs": {},
    },
    "ChaoShenMIEstimator": {
        "functional_str": ["chao_shen"],
        "needed_kwargs": {},
    },
    "ChaoWangJostMIEstimator": {
        "functional_str": ["chao_wang_jost"],
        "needed_kwargs": {},
    },
    "DiscreteMIEstimator": {
        "functional_str": ["discrete"],
        "needed_kwargs": {},
    },
    "GrassbergerMIEstimator": {
        "functional_str": ["grassberger"],
        "needed_kwargs": {},
    },
    "KernelMIEstimator": {
        "functional_str": ["kernel"],
        "needed_kwargs": {"bandwidth": 0.3, "kernel": "box"},
    },
    "KSGMIEstimator": {
        "functional_str": ["metric", "ksg"],
        "needed_kwargs": {},
    },
    "MillerMadowMIEstimator": {
        "functional_str": ["miller_madow", "mm"],
        "needed_kwargs": {},
    },
    "NsbMIEstimator": {
        "functional_str": ["nsb"],
        "needed_kwargs": {},
    },
    "OrdinalMIEstimator": {
        "functional_str": ["ordinal", "symbolic", "permutation"],
        "needed_kwargs": {"embedding_dim": 2},
    },
    "RenyiMIEstimator": {
        "functional_str": ["renyi"],
        "needed_kwargs": {"alpha": 1.5},
    },
    "ShrinkMIEstimator": {
        "functional_str": ["shrink"],
        "needed_kwargs": {},
    },
    "TsallisMIEstimator": {
        "functional_str": ["tsallis"],
        "needed_kwargs": {"q": 2.0},
    },
    "ZhangMIEstimator": {
        "functional_str": ["zhang"],
        "needed_kwargs": {},
    },
}

CMI_APPROACHES = {
    "AnsbCMIEstimator": {
        "functional_str": ["ansb"],
        "needed_kwargs": {},
    },
    "BayesCMIEstimator": {
        "functional_str": ["bayes"],
        "needed_kwargs": {"alpha": "jeffrey"},
    },
    "BonachelaCMIEstimator": {
        "functional_str": ["bonachela"],
        "needed_kwargs": {},
    },
    "ChaoShenCMIEstimator": {
        "functional_str": ["chao_shen"],
        "needed_kwargs": {},
    },
    "ChaoWangJostCMIEstimator": {
        "functional_str": ["chao_wang_jost"],
        "needed_kwargs": {},
    },
    "DiscreteCMIEstimator": {
        "functional_str": ["discrete"],
        "needed_kwargs": {},
    },
    "GrassbergerCMIEstimator": {
        "functional_str": ["grassberger"],
        "needed_kwargs": {},
    },
    "KernelCMIEstimator": {
        "functional_str": ["kernel"],
        "needed_kwargs": {"bandwidth": 0.3, "kernel": "box"},
    },
    "KSGCMIEstimator": {
        "functional_str": ["metric", "ksg"],
        "needed_kwargs": {},
    },
    "MillerMadowCMIEstimator": {
        "functional_str": ["miller_madow", "mm"],
        "needed_kwargs": {},
    },
    "NsbCMIEstimator": {
        "functional_str": ["nsb"],
        "needed_kwargs": {},
    },
    "OrdinalCMIEstimator": {
        "functional_str": ["ordinal", "symbolic", "permutation"],
        "needed_kwargs": {"embedding_dim": 2},
    },
    "RenyiCMIEstimator": {
        "functional_str": ["renyi"],
        "needed_kwargs": {"alpha": 1.5},
    },
    "ShrinkCMIEstimator": {
        "functional_str": ["shrink"],
        "needed_kwargs": {},
    },
    "TsallisCMIEstimator": {
        "functional_str": ["tsallis"],
        "needed_kwargs": {"q": 2.0},
    },
    "ZhangCMIEstimator": {
        "functional_str": ["zhang"],
        "needed_kwargs": {},
    },
}

TE_APPROACHES = {
    "AnsbTEEstimator": {
        "functional_str": ["ansb"],
        "needed_kwargs": {},
    },
    "BayesTEEstimator": {
        "functional_str": ["bayes"],
        "needed_kwargs": {"alpha": "jeffrey"},
    },
    "BonachelaTEEstimator": {
        "functional_str": ["bonachela"],
        "needed_kwargs": {},
    },
    "ChaoShenTEEstimator": {
        "functional_str": ["chao_shen"],
        "needed_kwargs": {},
    },
    "ChaoWangJostTEEstimator": {
        "functional_str": ["chao_wang_jost"],
        "needed_kwargs": {},
    },
    "DiscreteTEEstimator": {
        "functional_str": ["discrete"],
        "needed_kwargs": {},
    },
    "GrassbergerTEEstimator": {
        "functional_str": ["grassberger"],
        "needed_kwargs": {},
    },
    "KernelTEEstimator": {
        "functional_str": ["kernel"],
        "needed_kwargs": {"bandwidth": 0.3, "kernel": "box"},
    },
    "KSGTEEstimator": {
        "functional_str": ["metric", "ksg"],
        "needed_kwargs": {},
    },
    "MillerMadowTEEstimator": {
        "functional_str": ["miller_madow", "mm"],
        "needed_kwargs": {},
    },
    "NsbTEEstimator": {
        "functional_str": ["nsb"],
        "needed_kwargs": {},
    },
    "OrdinalTEEstimator": {
        "functional_str": ["ordinal", "symbolic", "permutation"],
        "needed_kwargs": {"embedding_dim": 2},
    },
    "RenyiTEEstimator": {
        "functional_str": ["renyi"],
        "needed_kwargs": {"alpha": 1.5},
    },
    "ShrinkTEEstimator": {
        "functional_str": ["shrink"],
        "needed_kwargs": {},
    },
    "TsallisTEEstimator": {
        "functional_str": ["tsallis"],
        "needed_kwargs": {"q": 2.0},
    },
    "ZhangTEEstimator": {
        "functional_str": ["zhang"],
        "needed_kwargs": {},
    },
}


CTE_APPROACHES = {
    "AnsbCTEEstimator": {
        "functional_str": ["ansb"],
        "needed_kwargs": {},
    },
    "BayesCTEEstimator": {
        "functional_str": ["bayes"],
        "needed_kwargs": {"alpha": "jeffrey"},
    },
    "BonachelaCTEEstimator": {
        "functional_str": ["bonachela"],
        "needed_kwargs": {},
    },
    "ChaoShenCTEEstimator": {
        "functional_str": ["chao_shen"],
        "needed_kwargs": {},
    },
    "ChaoWangJostCTEEstimator": {
        "functional_str": ["chao_wang_jost"],
        "needed_kwargs": {},
    },
    "DiscreteCTEEstimator": {
        "functional_str": ["discrete"],
        "needed_kwargs": {},
    },
    "GrassbergerCTEEstimator": {
        "functional_str": ["grassberger"],
        "needed_kwargs": {},
    },
    "KernelCTEEstimator": {
        "functional_str": ["kernel"],
        "needed_kwargs": {"bandwidth": 0.3, "kernel": "box"},
    },
    "KSGCTEEstimator": {
        "functional_str": ["metric", "ksg"],
        "needed_kwargs": {},
    },
    "MillerMadowCTEEstimator": {
        "functional_str": ["miller_madow", "mm"],
        "needed_kwargs": {},
    },
    "NsbCTEEstimator": {
        "functional_str": ["nsb"],
        "needed_kwargs": {},
    },
    "OrdinalCTEEstimator": {
        "functional_str": ["ordinal", "symbolic", "permutation"],
        "needed_kwargs": {"embedding_dim": 2},
    },
    "RenyiCTEEstimator": {
        "functional_str": ["renyi"],
        "needed_kwargs": {"alpha": 1.5},
    },
    "ShrinkCTEEstimator": {
        "functional_str": ["shrink"],
        "needed_kwargs": {},
    },
    "TsallisCTEEstimator": {
        "functional_str": ["tsallis"],
        "needed_kwargs": {"q": 2.0},
    },
    "ZhangCTEEstimator": {
        "functional_str": ["zhang"],
        "needed_kwargs": {},
    },
}


NOT_NORMALIZABLE = [
    "discrete",
    "miller_madow",
    "mm",
    "ordinal",
    "symbolic",
    "permutation",
    "ansb",
    "bayes",
    "bonachela",
    "chao_shen",
    "chao_wang_jost",
    "grassberger",
    "nsb",
    "shrink",
    "zhang",
]
HAVE_LOCAL_VALS = [
    "discrete",
    "shrink",
    "grassberger",
    "kernel",
    "metric",
    "ksg",
    "miller_madow",
    "mm",
    "ordinal",
    "symbolic",
    "permutation",
    "shrink",
    "zhang",
]
FILTER_WARNINGS = ["ansb", "chao_wang_jost", "shrink"]


@pytest.fixture(
    scope="function",
)
def default_rng():
    """A random number generator."""
    return rng(seed=798)


@pytest.fixture(autouse=True, scope="session")
def activate_debug_logging():
    """Activate debug logging for all tests."""
    Config.set_log_level("DEBUG")


@pytest.fixture(
    scope="session",
    params=ENTROPY_APPROACHES.keys(),
)
def entropy_estimator(request):
    """A fixture that yields entropy estimator classes, with specific kwargs for one."""
    return getattr(entropy, request.param), ENTROPY_APPROACHES[request.param][
        "needed_kwargs"
    ]


entropy_approach_kwargs = [
    (elem["functional_str"][i], elem["needed_kwargs"])
    for elem in ENTROPY_APPROACHES.values()
    for i in range(len(elem["functional_str"]))
]


@pytest.fixture(
    scope="session",
    params=entropy_approach_kwargs,
    ids=[eak[0] for eak in entropy_approach_kwargs],
)
def entropy_approach(request):
    """A fixture that yields a tuple of (approach_str, needed_kwargs)."""
    return request.param


@pytest.fixture(
    scope="session",
    params=MI_APPROACHES.keys(),
)
def mi_estimator(request):
    """A fixture that yields mutual information estimator classes."""
    return getattr(mutual_information, request.param), MI_APPROACHES[request.param][
        "needed_kwargs"
    ]


mi_approach_kwargs = [
    (elem["functional_str"][i], elem["needed_kwargs"])
    for elem in MI_APPROACHES.values()
    for i in range(len(elem["functional_str"]))
]


@pytest.fixture(
    scope="session",
    params=mi_approach_kwargs,
    ids=[mak[0] for mak in mi_approach_kwargs],
)
def mi_approach(request):
    """A fixture that yields a tuple of (approach_str, needed_kwargs)."""
    return request.param


@pytest.fixture(
    scope="session",
    params=CMI_APPROACHES.keys(),
)
def cmi_estimator(request):
    """A fixture that yields conditional mutual information estimator classes."""
    return getattr(mutual_information, request.param), CMI_APPROACHES[request.param][
        "needed_kwargs"
    ]


cmi_approach_kwargs = [
    (elem["functional_str"][i], elem["needed_kwargs"])
    for elem in CMI_APPROACHES.values()
    for i in range(len(elem["functional_str"]))
]


@pytest.fixture(
    scope="session",
    params=cmi_approach_kwargs,
    ids=[mak[0] for mak in cmi_approach_kwargs],
)
def cmi_approach(request):
    """A fixture that yields a tuple of (approach_str, needed_kwargs)."""
    return request.param


@pytest.fixture(
    scope="session",
    params=TE_APPROACHES.keys(),
)
def te_estimator(request):
    """A fixture that yields transfer entropy estimator classes."""
    return getattr(transfer_entropy, request.param), TE_APPROACHES[request.param][
        "needed_kwargs"
    ]


te_approach_kwargs = [
    (elem["functional_str"][i], elem["needed_kwargs"])
    for elem in TE_APPROACHES.values()
    for i in range(len(elem["functional_str"]))
]


@pytest.fixture(
    scope="session",
    params=te_approach_kwargs,
    ids=[tak[0] for tak in te_approach_kwargs],
)
def te_approach(request):
    """A fixture that yields a tuple of (approach_str, needed_kwargs)."""
    return request.param


@pytest.fixture(
    scope="session",
    params=CTE_APPROACHES.keys(),
)
def cte_estimator(request):
    """A fixture that yields conditional transfer entropy estimator classes."""
    return getattr(transfer_entropy, request.param), CTE_APPROACHES[request.param][
        "needed_kwargs"
    ]


cte_approach_kwargs = [
    (elem["functional_str"][i], elem["needed_kwargs"])
    for elem in CTE_APPROACHES.values()
    for i in range(len(elem["functional_str"]))
]


@pytest.fixture(
    scope="session",
    params=cte_approach_kwargs,
    ids=[tak[0] for tak in cte_approach_kwargs],
)
def cte_approach(request):
    """A fixture that yields a tuple of (approach_str, needed_kwargs)."""
    return request.param


@cache
def generate_autoregressive_series(rng_int, alpha, beta, gamma, length=1000, scale=10):
    # Initialize the series with zeros
    X = zeros(length)
    Y = zeros(length)
    generator = rng(rng_int)
    # Generate the series
    for i in range(length - 1):
        eta_X = generator.normal(loc=0, scale=scale)
        eta_Y = generator.normal(loc=0, scale=scale)
        X[i + 1] = alpha * X[i] + eta_X
        Y[i + 1] = beta * Y[i] + gamma * X[i] + eta_Y
    return X, Y


@cache
def generate_autoregressive_series_condition(
    rng_int, alpha: tuple, beta, gamma: tuple, length=1000, scale=10
):
    # Initialize the series with zeros
    X = zeros(length)
    Y = zeros(length)
    Z = zeros(length)
    generator = rng(rng_int)
    # Generate the series
    for i in range(length - 1):
        eta_X = generator.normal(loc=0, scale=scale)
        eta_Y = generator.normal(loc=0, scale=scale)
        eta_Z = generator.normal(loc=0, scale=scale)
        X[i + 1] = alpha[0] * X[i] + eta_X
        Z[i + 1] = alpha[1] * Z[i] + eta_Z
        Y[i + 1] = beta * Y[i] + gamma[0] * X[i] + gamma[1] * Z[i] + eta_Y

    return X, Y, Z


@cache
def discrete_random_variables(rng_int, prop_time=0, low=0, high=4, length=1000):
    """Generate two coupled discrete random variables.

    The first variable is a uniform random variable with values in [low, high-1].
    Variable 2 takes the highest bit of the previous value of Variable 1
    (if we take a 2 bit representation of variable 1)
    as its own lowest bit, then assigns its highest bit at random.

    So, the two should have ~1 bit of mutual information.
    """
    generator = rng(rng_int)
    X = generator.integers(low, high, length)
    Y = [0] * length
    for i in range(1, length):
        Y[i] = (X[i - 1 - prop_time] & 1) + (generator.integers(0, 2) << 1)
    return X, Y


@cache
def discrete_random_variables_shifted(rng_int, shift=1, low=0, high=2, length=1000):
    """
    Generate the source array filled with random integers in [low, high[.
    Create the destination array by circularly shifting the source array by `prop_time` index.

    A third random array is generated, uncoupled from the first two.

    Parameters
    ----------
    rng_int : int
        Random number generator seed.
    shift : int
        Time delay.
    low : int
        low
    high : int
        high
    length : int
        Length of the output arrays.

    Returns
    -------
    X : ndarray
        Source array.
    Y : ndarray
        Destination array.
    C : ndarray
        Control array.
    """
    generator = rng(rng_int)
    X = generator.integers(low, high, length)
    Y = roll(X, shift)  # e.g., roll([0, 1, 2, 3], 1) = [3, 0, 1, 2]
    return X, Y, generator.integers(low, high, length)


@cache
def discrete_random_variables_condition(rng_int, low=0, high=4, length=1000):
    """Generate three coupled discrete random variables.

    The first variable is a uniform random variable with values in [low, high-1].
    Variable 2 takes the highest bit of the previous value of Variable 1
    (if we take a 2 bit representation of variable 1)
    as its own lowest bit, then assigns its highest bit at random.
    Variable 3 takes the lowest bit of the previous value of Variable 1
    and the highest bit of the previous value of Variable 2.
    """
    generator = rng(rng_int)
    X = generator.integers(low, high, length)
    Y = [0] * length
    Z = [0] * length
    for i in range(1, length):
        Y[i] = (X[i - 1] & 1) + (generator.integers(0, 2) << 1)
        Z[i] = (X[i - 1] & 1) + (Y[i - 1] & 2)
    return X, Y, Z



"""Data structures and containers for the infomeasure package."""

from dataclasses import dataclass, field

from numpy import ndarray, unique, asarray, percentile, integer, issubdtype


@dataclass(frozen=True)
class DiscreteData:
    """Container for discrete random variable data.

    Attributes
    ----------
    uniq : ndarray
        Array of unique values in the discrete data.
    counts : ndarray
        Array of counts for each unique value.
    data : ndarray, optional
        Original data array. Defaults to None.
    N : int, optional
        Total number of samples (length of data). Defaults to None.
    K : int, optional
        Number of unique values (len(uniq)). Defaults to None.
    """

    uniq: ndarray
    counts: ndarray
    N: int = field(init=False)
    K: int = field(init=False)
    data: ndarray = None

    @classmethod
    def from_data(cls, data: ndarray) -> "DiscreteData":
        """Create a DiscreteData object from a data array.

        Parameters
        ----------
        data : ndarray
            Raw data array to analyse.

        Returns
        -------
        DiscreteData
            New instance with computed unique values and counts.
        """
        data = asarray(data)
        if len(data) == 0:
            raise ValueError(
                f"`data` must not be empty, got {data} with shape {data.shape} instead."
            )
        uniq, counts = unique(data, return_counts=True)
        return cls(uniq=uniq, counts=counts, data=data)

    @classmethod
    def from_counts(cls, uniq: ndarray, counts: ndarray) -> "DiscreteData":
        """
        Constructs a `DiscreteData` instance from unique values and their counts.

        This class method creates an instance of the `DiscreteData` class by
        specifying the unique values and their corresponding counts. The `data`
        attribute of the created instance is set to `None`.

        Parameters
        ----------
        uniq : ndarray
            An array of unique values.
        counts : ndarray
            An array of counts corresponding to the unique values.

        Returns
        -------
        DiscreteData
            A new instance of the `DiscreteData` class initialized with the
            given unique values and their counts.
        """
        return cls(uniq=asarray(uniq), counts=asarray(counts))

    def __post_init__(self):
        """Validate attributes and set N and K."""
        if len(self.uniq) != len(self.counts):
            raise ValueError("uniq and counts must have same length")
        if any(self.counts < 0):
            raise ValueError("counts must be non-negative")
        if self.counts.sum() == 0:
            raise ValueError("counts must sum to a positive value")
        if not issubdtype(self.counts.dtype, integer):
            raise ValueError("counts must be integers")

        # Set K (number of unique values)
        object.__setattr__(self, "K", len(self.uniq))

        # Set N if data available
        if self.data is not None:
            if len(self.data) == 0:
                raise ValueError("data must not be empty")
            # Consistency between data and counts
            if self.counts.sum() != len(self.data):
                raise ValueError("counts must sum to the length of data")
            object.__setattr__(self, "N", len(self.data))
        else:
            object.__setattr__(self, "N", int(self.counts.sum()))

    @property
    def probabilities(self) -> ndarray:
        """
        Computes and returns the probabilities by normalizing counts.

        The `probabilities` property calculates the probabilities as the ratio
        of counts to the total value `N`. This provides a normalized
        representation of counts as probabilities.

        Returns
        -------
        ndarray
            An array containing the probabilities, calculated by dividing
            counts by `N`.
        """
        return self.counts / self.N

    @property
    def distribution_dict(self) -> dict:
        """Dictionary mapping unique elements to their corresponding probabilities.

        Returns
        -------
        dict
            A dictionary where keys are unique elements and values are their
            corresponding probabilities.
        """
        return dict(zip(self.uniq, self.probabilities))


@dataclass(frozen=True)
class StatisticalTestResult:
    """Comprehensive statistical test result containing *p*-value, *t*-score,
    and confidence intervals.

    Attributes
    ----------
    p_value : float
        The *p*-value of the statistical test.
    t_score : float
        The *t*-score (standardized test statistic).
    test_values : ndarray
        The test values from permutation/bootstrap sampling.
    observed_value : float
        The observed value being tested.
    null_mean : float
        Mean of the null distribution (test values).
    null_std : float
        Standard deviation of the null distribution.
    n_tests : int
        Number of tests performed (permutations or bootstrap samples).
    method : str
        The statistical test method used ("permutation_test" or "bootstrap").
    """

    p_value: float
    t_score: float
    test_values: ndarray
    observed_value: float
    null_mean: float
    null_std: float
    n_tests: int
    method: str

    def percentile(self, q, method="linear"):
        """Compute the q-th percentile of the test values.

        This method wraps numpy's percentile function to compute percentiles
        of the test values from the statistical test.

        Parameters
        ----------
        q : array_like of float
            Percentage or sequence of percentages for the percentiles to compute.
            Values must be between 0 and 100 inclusive.
        method : str, optional
            Method to use for estimating the percentile. Default is "linear".
            See :py:func:`numpy.percentile` for available methods.

        Returns
        -------
        percentile : scalar or ndarray
            If `q` is a single percentile, returns a scalar.
            If multiple percentiles are given, returns an array.

        See Also
        --------
        numpy.percentile : Compute percentiles along specified axes.

        Notes
        -----
        For details on the method parameter, reference :py:func:`numpy.percentile`.

        Examples
        --------
        >>> result = estimator.statistical_test(n_tests=100,method="permutation_test")
        >>> result.percentile(50)  # Median
        >>> result.percentile([25, 75])  # Quartiles
        >>> result.percentile(95, method="nearest")  # 95th percentile with nearest method
        """
        return percentile(self.test_values, q, method=method)

    def confidence_interval(self, confidence_level, method="linear"):
        """Get confidence interval for the specified confidence level.

        This is a convenience function that converts a confidence level
        (e.g., 95 for 95% CI) to the appropriate percentile calls.

        Parameters
        ----------
        confidence_level : float
            Confidence level as a percentage (e.g., 95 for 95% CI).
            Must be between 0 and 100.
        method : str, optional
            Method to use for estimating the percentile. Default is "linear".
            See :py:func:`numpy.percentile` for available methods.

        Returns
        -------
        ndarray
            Array containing [lower_bound, upper_bound] of the confidence interval.

        Raises
        ------
        ValueError
            If confidence_level is not between 0 and 100.

        Examples
        --------
        >>> result = estimator.statistical_test(n_tests=100)
        >>> result.confidence_interval(95)  # 95% CI
        >>> result.confidence_interval(90, method="nearest")  # 90% CI with nearest method
        """
        # Validate confidence level
        if not 0 < confidence_level < 100:
            raise ValueError(
                f"Confidence level must be between 0 and 100, got {confidence_level}"
            )

        # Calculate percentiles: y = (100 - x) / 2
        y = (100 - confidence_level) / 2
        percentiles = [y, 100 - y]

        return self.percentile(percentiles, method=method)



"""Explicit Renyi entropy tests."""

import pytest
from numpy import inf

from infomeasure.estimators.entropy import (
    RenyiEntropyEstimator,
    DiscreteEntropyEstimator,
)


@pytest.mark.parametrize("k", [1, 2, 3, 5, 10])
@pytest.mark.parametrize("alpha", [0.5, 1.0, 1.5, 2.0, 3.0, 4.0])
def test_renyi_entropy(k, alpha, default_rng):
    """Test the Renyi entropy estimator by design."""
    data = default_rng.normal(0, 10, 1000)
    est = RenyiEntropyEstimator(data, k=k, alpha=alpha)
    if alpha == 1:
        est_discrete = DiscreteEntropyEstimator(data.astype(int))
        assert pytest.approx(est.result(), rel=0.1) == est_discrete.result()
    est.result()


@pytest.mark.parametrize(
    "data,k,alpha,expected",
    [
        ([1, 0, 1, 1, 1, 4, 23, 6, 1, -4, -3], 4, 1.0, 2.522422772),
        ([1, 2, 1, 2, 1, 2, 1, 2], 4, 1.1, 2.015828887),
        ([[0, 0], [1, 3], [2, 2], [3, 1], [4, 0]], 1, 1, 5.8842423067),
        ([[0, 0], [1, 3], [2, 2], [3, 1], [4, 0]], 2, 0.9, 5.297410751),
    ],
)
def test_renyi_entropy_explicit(data, k, alpha, expected):
    """Test the Renyi entropy estimator with specific values."""
    est = RenyiEntropyEstimator(data, k=k, alpha=alpha, base=2)
    assert est.result() == pytest.approx(expected)


@pytest.mark.parametrize("k", [0, -1, -10, None])
def test_renyi_entropy_invalid_k(k, default_rng):
    """Test the Renyi entropy estimator with invalid k."""
    data = list(range(10))
    with pytest.raises(ValueError):
        RenyiEntropyEstimator(data, k=k, alpha=1)


@pytest.mark.parametrize("alpha", [0, -1, -10, None])
def test_renyi_entropy_invalid_alpha(alpha, default_rng):
    """Test the Renyi entropy estimator with invalid alpha."""
    data = list(range(10))
    with pytest.raises(ValueError):
        RenyiEntropyEstimator(data, k=1, alpha=alpha)



"""
Explicit tests for the Kraskov-Stoegbauer-Grassberger (KSG) transfer entropy estimator.
"""

import pytest
from numpy import ndarray, inf

from tests.conftest import (
    generate_autoregressive_series,
    generate_autoregressive_series_condition,
)
from infomeasure.estimators.transfer_entropy import KSGTEEstimator, KSGCTEEstimator


@pytest.mark.parametrize(
    "rng_int,k,minkowski_p,base,expected",
    [
        (5, 4, 2, "e", -0.1464972645512768),
        (5, 4, 3, "e", 0.20521970071775472),
        (5, 4, inf, "e", 0.5959660770508142),
        (5, 16, 2, "e", -0.09411416419535731),
        (5, 16, 3, "e", 0.23892165815373245),
        (5, 16, inf, "e", 0.6034525014935584),
        (6, 4, 2, "e", -0.14070180426292878),
        (6, 4, 3, "e", 0.20704153706669481),
        (6, 4, inf, "e", 0.5660288107639446),
        (7, 4, 2, "e", -0.1333354231111923),
        (7, 4, 3, "e", 0.21280380426208417),
        (7, 4, inf, "e", 0.598763591023636),
        (7, 4, 2, 2.0, -0.19236235),
        (7, 4, 3, 10, 0.092419517),
        (7, 4, inf, 5, 0.372032736),
    ],
)
def test_ksg_te(rng_int, k, minkowski_p, base, expected):
    """Test the KSG transfer entropy estimator."""
    data_source, data_dest = generate_autoregressive_series(rng_int, 0.5, 0.6, 0.4)
    est = KSGTEEstimator(
        data_source,
        data_dest,
        k=k,
        minkowski_p=minkowski_p,
        noise_level=0,  # for reproducibility
        base=base,
    )
    assert isinstance(est.result(), float)
    assert est.result() == pytest.approx(expected)
    assert isinstance(est.local_vals(), ndarray)


@pytest.mark.parametrize(
    "rng_int,prop_time,step_size,src_hist_len,dest_hist_len,expected",
    [
        (5, 1, 1, 1, 1, 0.4979940181649),
        (6, 1, 1, 1, 1, 0.498845843022),
        (7, 1, 1, 1, 1, 0.513635125765),
        (5, 0, 1, 1, 1, 0.595966077050),
        (5, 1, 2, 1, 1, 0.51980907955),
        (5, 1, 3, 1, 1, 0.513413348695),
        (5, 1, 1, 2, 1, 0.914015718313),
        (5, 1, 1, 1, 2, 1.113892408337),
        (5, 1, 1, 2, 2, 1.580475709527),
    ],
)
def test_ksg_te_slicing(
    rng_int,
    prop_time,
    step_size,
    src_hist_len,
    dest_hist_len,
    expected,
):
    """Test the KSG transfer entropy estimator with slicing."""
    data_source, data_dest = generate_autoregressive_series(rng_int, 0.5, 0.6, 0.4)
    est = KSGTEEstimator(
        data_source,
        data_dest,
        prop_time=prop_time,
        step_size=step_size,
        src_hist_len=src_hist_len,
        dest_hist_len=dest_hist_len,
        noise_level=0,  # for reproducibility
    )
    assert isinstance(est.result(), float)
    assert est.result() == pytest.approx(expected)
    assert isinstance(est.local_vals(), ndarray)


@pytest.mark.parametrize(
    "rng_int,k,minkowski_p,base,expected",
    [
        (5, 4, 2, "e", -0.14824377373),
        (5, 4, 3, "e", 0.492061479471),
        (5, 4, inf, "e", 1.1736338791),
        (5, 16, 2, "e", -0.0629498022),
        (5, 16, 3, "e", 0.5019661670),
        (5, 16, inf, "e", 1.1344990340),
        (6, 4, 2, "e", -0.1120271189),
        (6, 4, 3, "e", 0.5267832946),
        (6, 4, inf, "e", 1.2010598113),
        (7, 4, 2, "e", -0.17448870909),
        (7, 4, 3, "e", 0.4594946023),
        (7, 4, inf, "e", 1.1573234261),
        (7, 4, 2, 2.0, -0.25173399),
        (7, 4, 3, 10, 0.199555970),
        (7, 4, inf, 5.0, 0.7190854),
    ],
)
def test_ksg_cte(rng_int, k, minkowski_p, base, expected):
    """Test the conditional KSG transfer entropy estimator."""
    data_source, data_dest, data_cond = generate_autoregressive_series_condition(
        rng_int, alpha=(0.5, 0.1), beta=0.6, gamma=(0.4, 0.2)
    )
    est = KSGCTEEstimator(
        data_source,
        data_dest,
        cond=data_cond,
        k=k,
        minkowski_p=minkowski_p,
        noise_level=0,  # for reproducibility
        base=base,
    )
    assert isinstance(est.result(), float)
    assert est.result() == pytest.approx(expected)
    assert isinstance(est.local_vals(), ndarray)


@pytest.mark.parametrize(
    "rng_int,step_size,src_hist_len,dest_hist_len,cond_hist_len,expected",
    [
        (5, 1, 1, 1, 1, 1.17363387),
        (6, 1, 1, 1, 1, 1.20105981),
        (7, 1, 1, 1, 1, 1.15732342),
        (5, 1, 1, 1, 1, 1.17363387),
        (5, 2, 1, 1, 1, 1.11722357),
        (5, 3, 1, 1, 1, 1.11096789),
        (5, 1, 2, 1, 1, 1.65521713),
        (5, 1, 1, 2, 1, 1.75221232),
        (5, 1, 1, 1, 2, 1.72224612),
        (5, 1, 2, 2, 1, 2.17620552),
        (5, 1, 2, 2, 2, 2.58013058),
    ],
)
def test_ksg_cte_slicing(
    rng_int,
    step_size,
    src_hist_len,
    dest_hist_len,
    cond_hist_len,
    expected,
):
    """Test the conditional KSG transfer entropy estimator with slicing."""
    data_source, data_dest, data_cond = generate_autoregressive_series_condition(
        rng_int, alpha=(0.5, 0.1), beta=0.6, gamma=(0.4, 0.2)
    )
    est = KSGCTEEstimator(
        data_source,
        data_dest,
        cond=data_cond,
        step_size=step_size,
        src_hist_len=src_hist_len,
        dest_hist_len=dest_hist_len,
        cond_hist_len=cond_hist_len,
        noise_level=0,  # for reproducibility
    )
    assert isinstance(est.result(), float)
    assert est.result() == pytest.approx(expected)
    assert isinstance(est.local_vals(), ndarray)


@pytest.mark.parametrize(
    "rng_int,method,p_te,p_cte",
    [
        (1, "permutation_test", 0.0, 0.0),
        (1, "bootstrap", 0.0, 0.0),
        (2, "permutation_test", 0.0, 0.0),
        (2, "bootstrap", 0.0, 0.0),
        (3, "permutation_test", 0.0, 0.0),
        (4, "permutation_test", 0.0, 0.0),
    ],
)
def test_ksg_te_statistical_test(rng_int, method, p_te, p_cte):
    """Test the KSG TE for p-values. Fix rng."""
    data_source, data_dest, data_cond = generate_autoregressive_series_condition(
        rng_int, alpha=(0.5, 0.1), beta=0.6, gamma=(0.4, 0.2)
    )
    est_te_xy = KSGTEEstimator(
        data_source, data_dest, k=4, minkowski_p=inf, noise_level=0, base=2, seed=8
    )
    est_cte_xy = KSGCTEEstimator(
        data_source,
        data_dest,
        cond=data_cond,
        k=4,
        minkowski_p=inf,
        noise_level=0,
        base=2,
        seed=8,
    )
    test = est_te_xy.statistical_test(method=method, n_tests=50)
    assert test.p_value == pytest.approx(p_te)
    test = est_cte_xy.statistical_test(method=method, n_tests=50)
    assert test.p_value == pytest.approx(p_cte)


@pytest.mark.parametrize(
    "rng_int,method,eff_te,eff_cte",
    [
        (1, "permutation_test", 0.11228866688671651, 0.0),
        (1, "bootstrap", 0.10695252204417915, 0.0),
        (2, "permutation_test", 0.04533644468841158, 0.0),
        (2, "bootstrap", 0.06294198770734777, 0.0),
        (3, "permutation_test", 0.06633458735803954, 0.0),
        (4, "permutation_test", 0.1011367204218852, 0.0),
    ],
)
def test_ksg_te_effective_val(rng_int, method, eff_te, eff_cte):
    """Test the KSG transfer entropy for effective values. Fix rng."""
    data_source, data_dest, data_cond = generate_autoregressive_series_condition(
        rng_int, alpha=(0.5, 0.1), beta=0.6, gamma=(0.4, 0.2)
    )
    est_te_xy = KSGTEEstimator(
        data_source, data_dest, k=4, minkowski_p=inf, noise_level=0, base="e", seed=8
    )
    est_cte_xy = KSGCTEEstimator(
        data_source,
        data_dest,
        cond=data_source,
        k=4,
        minkowski_p=inf,
        noise_level=0,
        base="e",
        seed=8,
    )
    assert est_te_xy.effective_val(method=method) == pytest.approx(eff_te)
    assert est_cte_xy.effective_val(method=method) == pytest.approx(eff_cte)



"""Module for the discrete Grassberger entropy estimator."""

from numpy import asarray, log
from scipy.special import digamma

from infomeasure.estimators.base import DiscreteHEstimator
from ...utils.exceptions import TheoreticalInconsistencyError


class GrassbergerEntropyEstimator(DiscreteHEstimator):
    r"""Discrete Grassberger entropy estimator.

    .. math::

        \hat{H}_{\text{Gr88}} = \sum_i \frac{n_i}{H} \left(\log(N) - \psi(n_i) - \frac{(-1)^{n_i}}{n_i + 1}  \right)

    :math:`\hat{H}_{\text{Gr88}}` is the Grassberger entropy,
    where :math:`n_i` are the counts,
    :math:`H` is the total number of observations :math:`N`,
    and :math:`\psi` is the digamma function
    :cite:p:`grassbergerFiniteSampleCorrections1988,grassbergerEntropyEstimatesInsufficient2008`.

    Attributes
    ----------
    *data : array-like
        The data used to estimate the entropy.
    """

    def _simple_entropy(self):
        """Calculate the Grassberger entropy of the data.

        Returns
        -------
        float
            The calculated entropy.
        """

        # Create a mapping from unique values to their counts
        count_dict = dict(zip(self.data[0].uniq, self.data[0].counts))

        # Vectorized calculation of local values
        n_i = asarray([count_dict[val] for val in self.data[0].data])
        local_values = log(self.data[0].N) - digamma(n_i) - ((-1) ** n_i) / (n_i + 1)

        # Convert to the requested base if needed
        if self.base != "e":
            local_values /= log(self.base)

        return local_values

    def _cross_entropy(self) -> float:
        """Calculate cross-entropy between two distributions.

        Raises
        ------
        TheoreticalInconsistencyError
            Cross-entropy is not theoretically sound for Grassberger estimator
            due to conceptual mismatch between bias correction and cross-entropy.
        """
        raise TheoreticalInconsistencyError(
            "Cross-entropy is not implemented for Grassberger estimator. "
            "The Grassberger correction is designed for bias correction in entropy "
            "estimation using count-based corrections, but cross-entropy mixes "
            "probabilities from one distribution with corrections from another, "
            "creating a theoretical inconsistency."
        )



"""Tests for the ordinal / permutation utility functions."""

from itertools import permutations

import pytest
from numpy import (
    array,
    ndarray,
    uint8,
    uint16,
    arange,
    uint32,
    uint64,
    apply_along_axis,
    iinfo,
    array_equal,
)
from numpy.testing import assert_array_equal
from scipy.special import factorial

from infomeasure.estimators.utils.ordinal import (
    symbolize_series,
    permutation_to_integer,
    reduce_joint_space,
)


@pytest.mark.parametrize(
    "series, embedding_dim, step_size, expected",
    [
        ([1], 1, 1, [[0]]),
        ([1, 2], 2, 1, [[0, 1]]),
        ([1, 2, 3], 2, 1, [[0, 1]] * 2),
        ([1, 2, 3, 4], 2, 1, [[0, 1]] * 3),
        ([1, 2, 1, 4], 2, 1, [[0, 1], [1, 0], [0, 1]]),
        ([1, 2, 3, 4, 5, 6, 7, 8, 9], 2, 1, [[0, 1]] * 8),
        ([-4, 2, 8, 3], 3, 1, [[0, 1, 2], [0, 2, 1]]),
        ([-4, 2, 8, 1], 3, 1, [[0, 1, 2], [2, 0, 1]]),
        ([0, 0, -1, 1, -2, 2, -3, 3], 2, 2, [[1, 0], [0, 1]] * 3),
        ([0, 0, -1, 1, -2, 2, -3, 3], 2, 4, [[1, 0], [0, 1]] * 2),
        ([0, 0, -1, 1, -2, 2, -3, 3], 2, 6, [[1, 0], [0, 1]]),
        (
            [-1e-6, 0, -1, 1, -2, 2, -3, 3],
            6,
            1,
            [[4, 2, 0, 1, 3, 5], [5, 3, 1, 0, 2, 4], [4, 2, 0, 1, 3, 5]],
        ),
        ([3.2, 2.1, -1.2, -3.4, 0.0], 2, 1, [[1, 0]] * 3 + [[0, 1]]),
        (["a", "b", "c", "d"], 2, 1, [[0, 1]] * 3),
    ],
)
def test_symbolize_series(series, embedding_dim, step_size, expected):
    """Test the symbolize_series function."""
    series = array(series)
    patterns = symbolize_series(series, embedding_dim, step_size)
    assert_array_equal(patterns, array(expected))


@pytest.mark.parametrize(
    "series, embedding_dim, expected",
    [
        ([1], 1, [0]),
        ([1, 2], 2, [0]),
        ([2, 1], 2, [1]),
        ([2, 1, 2], 2, [1, 0]),
        ([1, 2, 1], 2, [0, 1]),
        ([1, 2, 3, 4], 2, [0, 0, 0]),
        ([1, 2, 3], 3, [0]),
        ([1, 2, 3, 4], 3, [0, 0]),
        ([1, 2, 3, 4, 5, -3], 5, [0, 96]),
        ([1, 2, 3, 4, 5, 6], 6, [0]),
    ],
)
def test_symbolize_series_to_int(series, embedding_dim, expected):
    """Test the symbolize_series function with to_int=True."""
    series = array(series)
    patterns = symbolize_series(series, embedding_dim, to_int=True)
    assert_array_equal(patterns, array(expected))
    assert patterns.dtype == uint8 if embedding_dim < 6 else uint16


@pytest.mark.parametrize("data_len", [10, int(1e3)])
@pytest.mark.parametrize("embedding_dim", [2, 3, 4])
@pytest.mark.parametrize("step_size", [1, 2, 3])
@pytest.mark.parametrize("to_int", [True, False])
def test_symbolize_series_programmatic(
    data_len, embedding_dim, step_size, to_int, default_rng
):
    """Test the symbolize_series function programmatically."""
    series = default_rng.normal(size=data_len)
    symbols = symbolize_series(series, embedding_dim, step_size, to_int=to_int)
    assert isinstance(symbols, ndarray)
    # if to_int: False, check that shape is correct
    if not to_int:
        assert symbols.shape == (
            data_len - (embedding_dim - 1) * step_size,
            embedding_dim,
        )
    # if to_int: True, check that shape is correct
    else:
        assert symbols.shape == (data_len - (embedding_dim - 1) * step_size,)
        assert symbols.dtype == uint8


@pytest.mark.parametrize(
    "series, embedding_dim, step_size, error",
    [
        ([], 2, 1, AttributeError),  # input is list, not ndarray
        (tuple([1]), 2, 1, AttributeError),  # input is tuple, not ndarray
        (array([1]), 1.0, 1, TypeError),  # embedding_dim is float
        (array([1]), 1, 1.0, TypeError),  # step_size is float
        (array([1]), 0, 1, ValueError),  # embedding_dim is < 1
        (array([1]), -4, 1, ValueError),  # embedding_dim is < 1
        (array([1]), 1, 0, ValueError),  # step_size is < 1
        (array([1]), 1, -3, ValueError),  # step_size is < 1
        (array([]), 1, 1, ValueError),  # empty series
        (array([1, 2]), 3, 1, ValueError),  # series is too small for embedding_dim
        (
            array([1, 2]),
            2,
            2,
            ValueError,
        ),  # series is too small for step_size+embedding_dim
    ],
)
def test_symbolize_series_error(series, embedding_dim, step_size, error):
    """Test the symbolize_series function with an error."""
    with pytest.raises(error):
        symbolize_series(series, embedding_dim, step_size)


@pytest.mark.parametrize(
    "patt, expected, dtype",
    [
        ([0, 1], 0, uint8),
        ([1, 0], 1, uint8),
        ([0, 1, 2], 0, uint8),
        ([2, 1, 0], 5, uint8),
        ([1, 0, 2], 2, uint8),
        ([2, 0, 1], 4, uint8),
        ([0, 1], 0, uint16),
        ([1, 0], 1, uint16),
        ([0, 1, 2], 0, uint16),
        ([2, 1, 0], 5, uint16),
        ([1, 0, 2], 2, uint16),
        ([2, 0, 1], 4, uint16),
    ],
)
def test_permutation_to_integer(patt, expected, dtype):
    """Test the permutation_to_integer function."""
    result = permutation_to_integer(array(patt), dtype=dtype)
    assert result == expected
    assert result.dtype == dtype


@pytest.mark.parametrize("patt_len", [1, 2, 5, 6, 8, 9, 12, 13, 20])
@pytest.mark.parametrize("dtype", [uint8, uint16, uint32, uint64])
def test_permutation_to_integer_min_max_patt(patt_len, dtype):
    """Test the permutation_to_integer function with min and max permutations."""
    patt = arange(patt_len)
    result = permutation_to_integer(patt, dtype=dtype)
    assert result == 0
    # Max value should be factorial(patt_len) - 1
    result_inv = permutation_to_integer(patt[::-1], dtype=dtype)
    if (
        # 5! < 2**8
        patt_len < 6  # 2**8 < 6! < 2**16
        or (patt_len < 9 and dtype == uint16)  # 2**16 < 9! < 2**32
        or (patt_len < 13 and dtype == uint32)  # 2**32 < 13! < 2**64
        or (patt_len < 21 and dtype == uint64)  # 2**64 < 21!
    ):  # No overflow
        assert result_inv == factorial(patt_len) - 1
    else:  # Overflow
        assert result_inv != factorial(patt_len) - 1

    assert result.dtype == dtype
    assert result_inv.dtype == dtype


def test_permutation_to_integer_embedding_dim_too_large():
    """Test the permutation_to_integer function with an embedding_dim
    that is too large."""
    with pytest.raises(ValueError):
        permutation_to_integer(arange(21))


@pytest.mark.parametrize("embedding_dim", [1, 2, 3, 4, 5, 6])
@pytest.mark.parametrize("dtype", [uint8, uint16, uint32, uint64])
def test_permutation_to_integer_uniqueness(embedding_dim, dtype):
    """Test the uniqueness property of the permutation_to_integer function."""
    # Get all possible permutations of [0, 1, 2,..., embedding_dim-1] without repetition
    perms = array(list(permutations(range(embedding_dim))))
    results = apply_along_axis(permutation_to_integer, 1, perms, dtype=dtype)
    assert len(results) == factorial(embedding_dim)
    assert results.dtype == dtype
    if (embedding_dim < 6) or (embedding_dim >= 6 and dtype != uint8):
        assert results.max() == factorial(embedding_dim) - 1
        assert results.min() == 0
        # Check that all results are unique
        assert len(set(results)) == len(results)
    else:  # Overflow
        assert results.max() == iinfo(dtype).max
        assert results.min() == 0
        # The results cannot be unique,
        # as the number of permutations is larger than the dtype can handle
        assert len(set(results)) != len(results)


@pytest.mark.parametrize(
    "data,expected",
    [
        # 1D: in=out
        ([4, 5, 4, 5, 6], [4, 5, 4, 5, 6]),
        ([1.0], [1.0]),
        (["a", "b", "a", "b", "c"], ["a", "b", "a", "b", "c"]),
        # 2D: reduce
        ([[2]], [0]),
        ([[-5], ["M"]], [0, 1]),
        ([[4], [5], [4], [5], [6]], [0, 1, 0, 1, 2]),
        ([[1, 2], [2, 3], [1, 2], [2, 3], [3, 4]], [0, 1, 0, 1, 2]),
        ([[9, 1, 6], [1, 2, 3], [4, 1, 7]], [2, 0, 1]),
        ([[1, 1], [2, 2], [3, 3]], [0, 1, 2]),
        ([[3, 3], [2, 2], [1, 1]], [2, 1, 0]),
    ],
)
def test_reduce_joint_space(data, expected):
    """Test the reduction of the joint space."""
    assert array_equal(reduce_joint_space(array(data)), array(expected))


@pytest.mark.parametrize(
    "data",
    [
        ([[[1, 2], [2, 3]], [[1, 2], [2, 3]], [[3, 4], [4, 5]]]),
        0,
        1.0,
        "a",
        (
            [
                [[[1, 2], [2, 3]], [[1, 2], [2, 3]], [[3, 4], [4, 5]]],
                [[[1, 2], [2, 3]], [[1, 2], [2, 3]], [[3, 4], [4, 5]]],
            ]
        ),
    ],
)
def test_reduce_joint_space_error(data):
    """Test the reduction of the joint space with wrong input dimensions."""
    with pytest.raises(ValueError):
        reduce_joint_space(array(data))



"""Tests for the unit_ball_volume module."""

import pytest
from numpy import pi

from infomeasure.estimators.utils.unit_ball_volume import unit_ball_volume


@pytest.mark.parametrize(
    "d, p, r, expected",
    [
        (1, 2, 1, 2),  # d = 1, p = 2, r = 1
        (2, 2, 1, pi),  # d = 2, p = 2, r = 1
        (3, 2, 1, 4 / 3 * pi),  # d = 3, p = 2, r = 1
        (2, float("inf"), 1, 4),  # d = 2, p = inf, r = 1
        (2, 1, 1, 2),  # d = 2, p = 1, r = 1
        (2, 2, 2, 4 * pi),  # d = 2, p = 2, r = 2
        (3, 2, 2, 4 / 3 * pi * 8),  # d = 3, p = 2, r = 2
        (3, float("inf"), 2, 64),  # d = 3, p = inf, r = 2
        (3, 1, 2, 32 / 3),  # d = 3, p = 1, r = 2
    ],
)
def test_unit_ball_volume(d, p, r, expected):
    assert pytest.approx(unit_ball_volume(d, r, p), rel=1e-9) == expected



"""Tests for discrete transfer entropy utility functions."""

import pytest
from numpy import array, log, log2, log10, ndarray, mean
from numpy.random import default_rng
from numpy.testing import assert_allclose

from infomeasure.estimators.utils.discrete_transfer_entropy import combined_te_form
from infomeasure.estimators.utils.te_slicing import te_observations, cte_observations


# Mock slice method for testing
def mock_slice_method_3_outputs(*data, **kwargs):
    """Mock slice method that returns 3 arrays (for TE without conditional)."""
    # Simple mock that returns the first 3 data arrays as-is
    return data[0][:10], data[1][:10], data[0][1:11]


def mock_slice_method_4_outputs(*data, **kwargs):
    """Mock slice method that returns 4 arrays (for CTE with conditional)."""
    # Simple mock that returns 4 arrays
    return data[0][:10], data[1][:10], data[0][1:11], data[2][:10]


def mock_slice_method_invalid(*data, **kwargs):
    """Mock slice method that returns invalid number of arrays."""
    return data[0][:10], data[1][:10]  # Only 2 arrays - should cause error


class TestCombinedTeForm:
    """Test the combined_te_form function."""

    def test_basic_functionality_with_te_observations(self):
        """Test basic functionality with te_observations slice method."""
        # Arrange: Create simple test data
        rng = default_rng(42)
        source = rng.integers(0, 4, 100)
        dest = rng.integers(0, 4, 100)

        # Act: Calculate transfer entropy
        result = combined_te_form(
            te_observations,
            source,
            dest,
            local=False,
            src_hist_len=2,
            dest_hist_len=2,
            step_size=1,
        )

        # Assert: Result should be a float
        assert isinstance(result, float)
        assert not isinstance(result, ndarray)

    def test_basic_functionality_with_cte_observations(self):
        """Test basic functionality with cte_observations slice method."""
        # Arrange: Create simple test data
        rng = default_rng(42)
        source = rng.integers(0, 4, 100)
        dest = rng.integers(0, 4, 100)
        cond = rng.integers(0, 4, 100)

        # Act: Calculate conditional transfer entropy
        result = combined_te_form(
            cte_observations,
            source,
            dest,
            cond,
            local=False,
            src_hist_len=2,
            dest_hist_len=2,
            cond_hist_len=2,
            step_size=1,
        )

        # Assert: Result should be a float
        assert isinstance(result, float)
        assert not isinstance(result, ndarray)

    def test_local_vs_global_calculation(self):
        """Test that local and global calculations return different types."""
        # Arrange: Create simple test data
        rng = default_rng(42)
        source = rng.integers(0, 4, 50)
        dest = rng.integers(0, 4, 50)

        # Act: Calculate both local and global
        global_result = combined_te_form(
            te_observations,
            source,
            dest,
            local=False,
            src_hist_len=1,
            dest_hist_len=1,
            step_size=1,
        )
        local_result = combined_te_form(
            te_observations,
            source,
            dest,
            local=True,
            src_hist_len=1,
            dest_hist_len=1,
            step_size=1,
        )

        # Assert: Global should be float, local should be array
        assert isinstance(global_result, float)
        assert isinstance(local_result, ndarray)
        assert local_result.ndim == 1
        assert len(local_result) > 0
        assert mean(local_result) == pytest.approx(global_result)

    @pytest.mark.parametrize(
        "log_func, base",
        [
            (log, "e"),
            (log2, 2),
            (log10, 10),
        ],
    )
    def test_different_log_functions(self, log_func, base):
        """Test with different logarithm functions."""
        # Arrange: Create simple test data
        rng = default_rng(42)
        source = rng.integers(0, 3, 50)
        dest = rng.integers(0, 3, 50)

        # Act: Calculate transfer entropy with different log functions
        result = combined_te_form(
            te_observations,
            source,
            dest,
            local=False,
            log_func=log_func,
            src_hist_len=1,
            dest_hist_len=1,
            step_size=1,
        )

        # Assert: Result should be a finite float
        assert isinstance(result, float)
        assert not (result != result)  # Check for NaN
        assert abs(result) != float("inf")  # Check for infinity

    def test_miller_madow_correction_applied(self):
        """Test that Miller-Madow correction changes the result."""
        # Arrange: Create simple test data
        rng = default_rng(42)
        source = rng.integers(0, 3, 100)
        dest = rng.integers(0, 3, 100)

        # Act: Calculate with and without Miller-Madow correction
        result_no_correction = combined_te_form(
            te_observations,
            source,
            dest,
            local=False,
            src_hist_len=2,
            dest_hist_len=2,
            step_size=1,
        )
        result_with_correction = combined_te_form(
            te_observations,
            source,
            dest,
            local=False,
            miller_madow_correction=2,
            src_hist_len=2,
            dest_hist_len=2,
            step_size=1,
        )

        # Assert: Results should be different
        assert result_no_correction != result_with_correction
        assert isinstance(result_with_correction, float)

    @pytest.mark.parametrize(
        "correction_value",
        [2, 10, "e", 2.0, 10.0],
    )
    def test_miller_madow_correction_different_bases(self, correction_value):
        """Test Miller-Madow correction with different base values."""
        # Arrange: Create simple test data
        rng = default_rng(42)
        source = rng.integers(0, 3, 50)
        dest = rng.integers(0, 3, 50)

        # Act: Calculate with Miller-Madow correction
        result = combined_te_form(
            te_observations,
            source,
            dest,
            local=False,
            miller_madow_correction=correction_value,
            src_hist_len=1,
            dest_hist_len=1,
            step_size=1,
        )

        # Assert: Result should be a finite float
        assert isinstance(result, float)
        assert not (result != result)  # Check for NaN
        assert abs(result) != float("inf")  # Check for infinity

    def test_mock_slice_method_3_outputs(self):
        """Test with mock slice method returning 3 outputs."""
        # Arrange: Create simple test data
        source = array([1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12])
        dest = array([2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13])

        # Act: Use mock slice method
        result = combined_te_form(
            mock_slice_method_3_outputs,
            source,
            dest,
            local=False,
        )

        # Assert: Should return a float result
        assert isinstance(result, float)

    def test_mock_slice_method_4_outputs(self):
        """Test with mock slice method returning 4 outputs."""
        # Arrange: Create simple test data
        source = array([1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12])
        dest = array([2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13])
        cond = array([3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14])

        # Act: Use mock slice method
        result = combined_te_form(
            mock_slice_method_4_outputs,
            source,
            dest,
            cond,
            local=False,
        )

        # Assert: Should return a float result
        assert isinstance(result, float)

    def test_invalid_slice_method_raises_error(self):
        """Test that invalid slice method output raises ValueError."""
        # Arrange: Create simple test data
        source = array([1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12])
        dest = array([2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13])

        # Act & Assert: Should raise ValueError
        with pytest.raises(ValueError, match="Invalid number of data arrays"):
            combined_te_form(
                mock_slice_method_invalid,
                source,
                dest,
                local=False,
            )

    def test_consistent_results_with_same_seed(self):
        """Test that results are consistent with the same random seed."""
        # Arrange: Create test data with same seed
        rng1 = default_rng(123)
        source1 = rng1.integers(0, 4, 50)
        dest1 = rng1.integers(0, 4, 50)

        rng2 = default_rng(123)
        source2 = rng2.integers(0, 4, 50)
        dest2 = rng2.integers(0, 4, 50)

        # Act: Calculate transfer entropy with same parameters
        result1 = combined_te_form(
            te_observations,
            source1,
            dest1,
            local=False,
            src_hist_len=1,
            dest_hist_len=1,
            step_size=1,
        )
        result2 = combined_te_form(
            te_observations,
            source2,
            dest2,
            local=False,
            src_hist_len=1,
            dest_hist_len=1,
            step_size=1,
        )

        # Assert: Results should be identical
        assert result1 == result2

    def test_local_calculation_with_miller_madow(self):
        """Test local calculation with Miller-Madow correction."""
        # Arrange: Create simple test data
        rng = default_rng(42)
        source = rng.integers(0, 3, 50)
        dest = rng.integers(0, 3, 50)

        # Act: Calculate local values with Miller-Madow correction
        result = combined_te_form(
            te_observations,
            source,
            dest,
            local=True,
            miller_madow_correction=2,
            src_hist_len=1,
            dest_hist_len=1,
            step_size=1,
        )

        # Assert: Should return array of local values
        assert isinstance(result, ndarray)
        assert result.ndim == 1
        assert len(result) > 0
        # All values should be finite
        assert all(not (val != val) for val in result)  # Check for NaN
        assert all(abs(val) != float("inf") for val in result)  # Check for infinity

    def test_conditional_transfer_entropy_with_miller_madow(self):
        """Test conditional transfer entropy with Miller-Madow correction."""
        # Arrange: Create simple test data
        rng = default_rng(42)
        source = rng.integers(0, 3, 100)
        dest = rng.integers(0, 3, 100)
        cond = rng.integers(0, 3, 100)

        # Act: Calculate CTE with Miller-Madow correction
        result = combined_te_form(
            cte_observations,
            source,
            dest,
            cond,
            local=False,
            miller_madow_correction="e",
            src_hist_len=1,
            dest_hist_len=1,
            cond_hist_len=1,
            step_size=1,
        )

        # Assert: Should return a finite float
        assert isinstance(result, float)
        assert not (result != result)  # Check for NaN
        assert abs(result) != float("inf")  # Check for infinity

    def test_slice_kwargs_passed_correctly(self):
        """Test that slice_kwargs are passed correctly to slice method."""
        # Arrange: Create simple test data
        rng = default_rng(42)
        source = rng.integers(0, 4, 100)
        dest = rng.integers(0, 4, 100)

        # Act: Calculate with specific slice parameters
        result = combined_te_form(
            te_observations,
            source,
            dest,
            local=False,
            src_hist_len=3,
            dest_hist_len=2,
            step_size=2,
        )

        # Assert: Should complete without error and return float
        assert isinstance(result, float)

    def test_edge_case_minimal_data(self):
        """Test with minimal data that still allows calculation."""
        # Arrange: Create minimal test data
        source = array([0, 1, 0, 1, 0, 1])
        dest = array([1, 0, 1, 0, 1, 0])

        # Act: Calculate with minimal parameters
        result = combined_te_form(
            te_observations,
            source,
            dest,
            local=False,
            src_hist_len=1,
            dest_hist_len=1,
            step_size=1,
        )

        # Assert: Should return a finite float
        assert isinstance(result, float)
        assert not (result != result)  # Check for NaN
        assert abs(result) != float("inf")  # Check for infinity

    def test_different_data_types(self):
        """Test with different data types (int, float, string)."""
        # Test with integer data
        source_int = array([1, 2, 3, 1, 2, 3, 1, 2])
        dest_int = array([3, 1, 2, 3, 1, 2, 3, 1])

        result_int = combined_te_form(
            te_observations,
            source_int,
            dest_int,
            local=False,
            src_hist_len=1,
            dest_hist_len=1,
            step_size=1,
        )

        # Test with string data
        source_str = array(["a", "b", "c", "a", "b", "c", "a", "b"])
        dest_str = array(["c", "a", "b", "c", "a", "b", "c", "a"])

        result_str = combined_te_form(
            te_observations,
            source_str,
            dest_str,
            local=False,
            src_hist_len=1,
            dest_hist_len=1,
            step_size=1,
        )

        # Assert: Both should return finite floats
        assert isinstance(result_int, float)
        assert isinstance(result_str, float)
        assert not (result_int != result_int)  # Check for NaN
        assert not (result_str != result_str)  # Check for NaN
